(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        29 July 2011
 *  o File name:   GenLib.sml
 *  o Description: generates a list of constant declarations.
 *)


structure GenLib :> GENLIB = struct

structure E  = Env
structure D  = Deps
structure T  = ListFormat
structure P  = ParserNuprlAscii
structure N  = NuprlTerms
structure EH = LibBase

structure MAP = BinaryMapFn(type ord_key = string val compare = String.compare)

type nuprl_so_types = (string * E.ity) list

type nuprl_scheme = string list                              (* bound type variables       *)
		    * (string * nuprl_so_types * E.ity) list (* arguments with their types *)
		    * E.ity                                  (* returned type              *)

val empty_nuprl_map =
    ref (MAP.empty : (N.parameter_value
		      * N.nuprl_term
		      * N.nuprl_term
		      * (N.nuprl_term * nuprl_scheme option) list) list MAP.map)

fun map2Lisp map =
    let val lst =
	    MAP.foldri (fn (id, entries, lst) => (*(obid, lhs, rhs, wfs)*)
			   id :: lst)
		       []
		       map
    in T.fmt {init  = "(",
	      final = ")",
	      sep   = " . ",
	      fmt   = fn x => x}
	     lst
    end

(*
fun exportMap2Lisp' output =
    let val output' = output ^ ".tmp"
	val stout   = TextIO.openOut output'
	val ()      =
	    MAP.appi (fn (id, entries) => (*(obid, lhs, rhs, wfs)*)
			 (TextIO.output (stout, id ^ ":\n");
			  foldl (fn ((obid, lhs, rhs, wfs), n) =>
				    foldl (fn ((wf, NONE), n) =>
					      let val st = "    type"     ^
							   Int.toString n ^
							   "?\n"
					      in (TextIO.output (stout, st); n+1)
					      end
					    | ((wf, SOME (bound, args, ity)), n) =>
					      let val ast =
						      if List.null args
						      then ""
						      else T.fmt {init  = "(",
								  final = ")",
								  sep   = "; ",
								  fmt   = fn (x, ity) => x ^ ": " ^ E.ppIty'' ity (fn x => "'" ^ x)}
								 args
						  val st ="    type"     ^
							  Int.toString n ^
							  ": constant "  ^
							  id             ^
							  " "            ^
							  ast            ^
							  " : "          ^
							  E.ppIty' ity   ^
							  "\n"
					      in (TextIO.output (stout, st); n+1)
					      end)
					  n
					  wfs)
				1
				entries;
			  TextIO.output (stout, "\n");
			  ()))
		     (getMap ())
	val _       = TextIO.closeOut stout
	val _       = OS.FileSys.rename {old = output', new = output}
    in ()
    end
*)

fun exportMap2Lisp map output =
    let fun ppItyTik ity = E.ppIty'' ity (fn x => "'" ^ x)
	fun ppSos sos =
	    if List.null sos
	    then ""
	    else T.fmt {init  = "[",
			final = "]",
			sep   = "; ",
			fmt   = fn (x, ity) => x ^ ": " ^ ppItyTik ity}
		       sos
	fun ppArgs args =
	    if List.null args
	    then ""
	    else T.fmt {init  = " (",
			final = ")",
			sep   = "; ",
			fmt   = fn (x, sos, ity) => x ^ ppSos sos ^ ": " ^ ppItyTik ity}
		       args
	fun toConstant id args ity =
	    "constant "  ^
	    id           ^
	    ppArgs args  ^
	    " : "        ^
	    ppItyTik ity ^
	    " ;;\n"
	val output' = output ^ ".tmp"
	val stout   = TextIO.openOut output'
	val ()      =
	    MAP.appi (fn (id, entries) => (*(obid, lhs, rhs, wfs)*)
			 if Substring.isSubstring "\\ " (Substring.full id)
			 then ()
			 else app (fn (obid, lhs, rhs, wfs) =>
				      app (fn (wf, NONE) => ()
					    | (wf, SOME (bound, args, ity)) =>
					      let val st = toConstant id args ity
					      in TextIO.output (stout, st)
					      end)
					  wfs)
				  entries)
		     map
	val _ = TextIO.closeOut stout
	val _ = OS.FileSys.rename {old = output', new = output}
    in ()
    end

(* 164, 186, 199, 206, 214, 273, 288 *)
(* pi1, pi2 *)

datatype SUB = POS | NEG

fun is_nuprl_type_term (N.TERM ((("universe", tag), params), [])) = true
  | is_nuprl_type_term _ = false

fun dest_nuprl_applys_term (N.TERM ((("apply", tag), params), [N.B_TERM ([], f), N.B_TERM ([], arg)])) =
    let val (h, args) = dest_nuprl_applys_term f
    in (h, args @ [arg])
    end
  | dest_nuprl_applys_term (term as N.TERM _) = (term, [])

fun nuprlType2Ity sub bound (N.TERM ((("variable", tag), [(v, "v")]), [])) =
    if List.exists (fn x => (x = v)) bound
    then SOME (E.ITYETV (v, D.dummy_label))
    else NONE
  | nuprlType2Ity sub bound (N.TERM ((("universe", tag), params), [])) =
    SOME (E.mk_type_type D.dummy_label)
  | nuprlType2Ity sub bound (N.TERM ((("list", tag), []), [N.B_TERM ([], t)])) =
    (case nuprlType2Ity sub bound t of
	 SOME ity => SOME (E.mk_type_list ity D.dummy_label)
       | NONE => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("bag", tag), []), [N.B_TERM ([], t)])) =
    (case nuprlType2Ity sub bound t of
	 SOME ity => SOME (E.mk_type_bag ity D.dummy_label)
       | NONE => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("eclass", tag), [lexp_param]), [N.B_TERM ([], info), N.B_TERM ([e, es], t)])) =
    (case nuprlType2Ity sub bound t of
	 SOME ity => SOME (E.mk_type_class ity D.dummy_label)
       | NONE => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("deq", tag), []), [N.B_TERM ([], t)])) =
    (case nuprlType2Ity sub bound t of
	 SOME ity => SOME (E.mk_type_deq ity D.dummy_label)
       | NONE => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("union", tag), []), [N.B_TERM ([], arg1), N.B_TERM ([], arg2)])) =
    (case (nuprlType2Ity sub bound arg1, nuprlType2Ity sub bound arg2) of
	 (SOME ity1, SOME ity2) => SOME (E.mk_type_disju (ity1, ity2) D.dummy_label)
       | _ => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("function", tag), []), [N.B_TERM ([], arg1), N.B_TERM ([""], arg2)])) =
    let val (sub1, sub2) = case sub of NEG => (POS, NEG) | POS => (NEG, POS)
    in case (nuprlType2Ity sub1 bound arg1, nuprlType2Ity sub2 bound arg2) of
	   (SOME ity1, SOME ity2) => SOME (E.mk_type_arrow (ity1, ity2) D.dummy_label)
	 | _ => NONE
    end
  | nuprlType2Ity sub bound (N.TERM ((("set", tag), []), [N.B_TERM ([], typ), N.B_TERM ([x], prop)])) =
    (case sub of
	 POS => nuprlType2Ity sub bound typ
       | NEG => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("product", tag), []), [N.B_TERM ([], arg1), N.B_TERM ([""], arg2)])) =
    (case (nuprlType2Ity sub bound arg1, nuprlType2Ity sub bound arg2) of
	 (SOME ity1, SOME ity2) => SOME (E.mk_type_tuple [ity1, ity2] D.dummy_label)
       | _ => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("nat", tag), []), [])) =
    (case sub of
	 POS => SOME (E.mk_type_int D.dummy_label)
       | NEG => NONE)
  | nuprlType2Ity sub bound (N.TERM ((("int",     tag), []), [])) = SOME (E.mk_type_int  D.dummy_label)
  | nuprlType2Ity sub bound (N.TERM ((("bool",    tag), []), [])) = SOME (E.mk_type_bool D.dummy_label)
  | nuprlType2Ity sub bound (N.TERM ((("atom",    tag), []), [])) = SOME (E.mk_type_atom D.dummy_label)
  | nuprlType2Ity sub bound (N.TERM ((("Id",      tag), []), [])) = SOME (E.mk_type_loc  D.dummy_label)
  | nuprlType2Ity sub bound (N.TERM ((("Message", tag), []), [])) = SOME (E.mk_type_msg  D.dummy_label)
  | nuprlType2Ity sub bound (N.TERM (((opid, tag), params), bterm)) = NONE

fun isSOApply [] term = SOME (term, [])
  | isSOApply vars (N.TERM ((("so_apply", tag), []), (N.B_TERM ([], term)) :: args)) =
    (let val lst = ListPair.zipEq (vars, args)
     in if List.all (fn (v, N.B_TERM ([], t)) =>
			N.is_nuprl_variable_term t
			andalso
			N.dest_variable t = v
		      | _ => false)
		    lst
	then SOME (term, vars)
	else NONE
     end handle ListPair.UnequalLengths => NONE)
  | isSOApply _ _ = NONE

fun nuprlType2ItyTerm id boundTypes boundIds (N.TERM (((opid, tag), params), bterms)) =
    if id = opid
    then foldr (fn (N.B_TERM (vars, term), SOME args) =>
		   (case isSOApply vars term of
			SOME (term, vars) => SOME ((term, vars) :: args)
		      | NONE => NONE)
		 | _ => NONE)
	       (SOME [])
	       bterms
    else NONE

fun nuprlType2ItyArg boundTypes boundIds (N.TERM ((("variable", tag), [(v, "v")]), [])) =
    List.find (fn (x, ity) => (x = v)) boundIds
  | nuprlType2ItyArg boundTypes boundIds (N.TERM ((("pair", tag), []), [N.B_TERM ([], term1), N.B_TERM ([], term2)])) =
    let val op1 = nuprlType2ItyArg boundTypes boundIds term1
	val op2 = nuprlType2ItyArg boundTypes boundIds term2
    in case (op1, op2) of
	   (SOME (v1, typ1), SOME (v2, typ2)) =>
	   SOME (v1 ^ v2, E.mk_type_tuple [typ1, typ2] D.dummy_label)
	 | _ => NONE
    end
  | nuprlType2ItyArg boundTypes boundIds (N.TERM (((opid, tag), params), bterms)) = NONE

fun soArgs [] (v, ity) args = SOME (v, args, ity)
  | soArgs (var :: vars) (v, ity) args =
    case E.destArrowType ity of
	SOME (t1, t2) => soArgs vars (v, t2) (args @ [(var, t1)])
      | NONE => NONE

fun nuprlType2ItyMember id boundTypes boundIds (N.TERM ((("member", tag), params), [N.B_TERM ([], typ), N.B_TERM ([], term)])) =
    let val (h, args) = dest_nuprl_applys_term term
    in case nuprlType2Ity POS boundTypes typ of
	   SOME ity =>
	   (case nuprlType2ItyTerm id boundTypes boundIds h of
		SOME args' =>
		let val tailop =
			foldr (fn (term, SOME ity) =>
				  (case nuprlType2ItyArg boundTypes boundIds term of
				       SOME (_, ity') => SOME (E.mk_type_arrow (ity', ity) D.dummy_label)
				     | NONE => NONE)
				| _ => NONE)
			      (SOME ity)
			      args
		in case tailop of
		       SOME itytail =>
		       foldr (fn ((term, vars), SOME (xs, itytail)) =>
				 (case nuprlType2ItyArg boundTypes boundIds term of
				      SOME pair =>
				      (case soArgs vars pair [] of
					   SOME x => SOME (x :: xs, itytail)
					 | NONE => NONE)
				    | NONE => NONE)
			       | _ => NONE)
			     (SOME ([], itytail))
			     args'
		     | NONE => NONE
		end
	      | NONE => NONE)
	 | NONE => NONE
    end
  | nuprlType2ItyMember id boundTypes boundIds (N.TERM ((("uimplies", tag), []), [N.B_TERM ([], A), N.B_TERM ([], B)])) =
    (* NOTE: Arg, that's not the best thing to do but we want to get
     * a shot at generating a type. *)
    nuprlType2ItyMember id boundTypes boundIds B
  | nuprlType2ItyMember _ _ _ (N.TERM (((opid, tag), params), bterms)) = NONE

fun nuprlType2ItyAll id boundTypes boundIds (N.TERM ((("all", tag), params), [N.B_TERM ([], T), N.B_TERM ([x], P)])) =
    if is_nuprl_type_term T
    then let val boundTypes' = x :: boundTypes
	 in nuprlType2ItyAll id boundTypes' boundIds P
	 end
    else (case nuprlType2Ity NEG boundTypes T of
	      SOME ity =>
	      let val boundIds' = (x, ity) :: boundIds
	      in nuprlType2ItyAll id boundTypes boundIds' P
	      end
	    | NONE => NONE)
  | nuprlType2ItyAll id boundTypes boundIds (N.TERM ((("uall", tag), params), [N.B_TERM ([], T), N.B_TERM ([x], P)])) =
    if is_nuprl_type_term T
    then let val boundTypes' = x :: boundTypes
	 in nuprlType2ItyAll id boundTypes' boundIds P
	 end
    else (case nuprlType2Ity NEG boundTypes T of
	      SOME ity =>
	      let val boundIds' = (x, ity) :: boundIds
	      in nuprlType2ItyAll id boundTypes boundIds' P
	      end
	    | NONE => NONE)
  | nuprlType2ItyAll id boundTypes boundIds (term as N.TERM (((opid, tag), params), bterms)) =
    (case nuprlType2ItyMember id boundTypes boundIds term of
	 SOME (args, ity) => SOME (boundTypes, args, ity)
       | NONE => NONE)

fun nuprlType2ItyAllWf id boundTypes boundIds (N.TERM (((opid, tag), []), [N.B_TERM ([], wf)])) =
    nuprlType2ItyAll id boundTypes boundIds wf
  | nuprlType2ItyAllWf _ _ _ _ = raise EH.Impossible "wrong format"

fun term2map (N.TERM (((id, "t"), [(obid, "o")]),
		      [N.B_TERM ([], lhs),
		       N.B_TERM ([], rhs),
		       N.B_TERM ([], wfs)])) =
    let val wfs' = map (fn wf =>
			   (case nuprlType2ItyAllWf id [] [] wf of
			       NONE => (wf, NONE)
			     | SOME (x as (ebound, args, ty)) => (wf, SOME x)))
		       (if N.is_nuprl_iwf_lemmas_term wfs
			then N.subterms wfs
			else raise EH.Impossible "term2map:wf_lemmas")
    in SOME (id, obid, lhs, rhs, wfs')
    end
  | term2map (N.TERM (((id, "t"), [(obid, "o")]),
		      [N.B_TERM ([], termof),
		       N.B_TERM ([], extract)])) = NONE
  (* We don't collect the extracts because this map is meant to collect
   * the wf lemmas to generate EventML constant declarations. *)
  | term2map term = (print (N.toStringTerm term);
		     raise EH.Impossible "wrong format")

fun b_terms2map [] = MAP.empty
  | b_terms2map ((N.B_TERM ([], term)) :: b_terms) =
    (case term2map term of
	 SOME (id, obid, lhs, rhs, wfs) =>
	 MAP.insert (b_terms2map b_terms, id, [(obid, lhs, rhs, wfs)])
       | NONE => b_terms2map b_terms)
  | b_terms2map _ = raise EH.Impossible "wrong format"

fun terms2map [] = MAP.empty
  | terms2map ((N.TERM (((opid, "t"), []), b_terms)) :: terms) =
    let val map1 = b_terms2map b_terms
	val map2 = terms2map terms
    in MAP.unionWithi (fn (id, list1, list2) =>
			  (*raise EH.Impossible (id ^ " occurs twice")*)
			  list1 @ list2)
		      (map1, map2)
    end
  | terms2map _ = raise EH.Impossible "wrong format"

type timer = {real : Timer.real_timer,
	      cpu  : Timer.cpu_timer}

fun startTimer () = {real = Timer.startRealTimer (),
		     cpu  = Timer.startCPUTimer ()}

fun getTime (timer : timer) = Timer.checkRealTimer (#real timer)

fun getMilliTime timer = Time.toMilliseconds (getTime timer)
fun getSecTime   timer = Time.toSeconds      (getTime timer)

fun cond_print true str = print str
  | cond_print false _ = ()

fun parseTerms input prt split = P.parse prt input split

fun exportTerms input outputop prt split =
    let val terms  = P.parse prt input split
	val _      = cond_print prt ("[start mapping]\n")
	val timer  = startTimer ()
	val map    = terms2map terms
	val _      = if prt
		     then app (fn t => cond_print prt (N.toStringTerm t ^ "\n")) terms
		     else ()
	val time   = getSecTime timer
	val _      = cond_print prt ("[end mapping: " ^ LargeInt.toString time ^ "s]\n")
	val _      =
	    case outputop of
		NONE => ()
	      | SOME output => exportMap2Lisp map output
	val _      = cond_print prt ("[export done]\n")
    in ()
    end

end
