(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   ToNuprl.sml
 *  o Description: Compiles EventML to Nuprl.
 *)


structure ToNuprl :> TONUPRL = struct

structure A  = Ast
structure E  = Env
structure D  = Deps
structure EN = Enum
structure T  = ListFormat
structure EH = LibBase
structure NT = NuprlTerms


val toStringTerm = NT.toStringTerm

(* ------ IDENTIFIERS ------ *)

structure IDS = A.IDS

fun get_all_ids_bterm (NT.B_TERM (variables, term)) =
    IDS.addList (get_all_ids term, variables)

and get_all_ids (term as NT.TERM (((opid, tag), params), bterms)) =
    let val set =
	    if NT.is_nuprl_variable_term term
	    then IDS.add (IDS.singleton opid, NT.dest_variable term)
	    else IDS.singleton opid
    in foldr (fn (bterm, set) => IDS.union (set, get_all_ids_bterm bterm))
	     set
	     bterms
    end

fun get_all_ids_list lst =
    let val some_ids =
	    ["bag-map", "bag-eq", "lambda", "pair", "eclass",
	     "normal-locally-programmable", "programmable",
	     "variable"]
    in foldr (fn (term, set) => IDS.union (set, get_all_ids term))
	     (IDS.addList (IDS.empty, some_ids))
	     lst
    end


(* ------ CONSTRUCTORS ------ *)

fun mk_nuprl_eta1_term v term =
    let val set = get_all_ids_list [term]
	val id  = A.newIdIdSet set v
	val nid = NT.mk_nuprl_variable_term id
	val app = NT.mk_nuprl_apply_term term nid
    in NT.mk_nuprl_lambda_term id term
    end

fun mk_nuprl_eta2_term v1 v2 term =
    let val set  = get_all_ids_list [term]
	val id1  = A.newIdIdSet set v1
	val id2  = A.newIdIdSet set v2
	val nid1 = NT.mk_nuprl_variable_term id1
	val nid2 = NT.mk_nuprl_variable_term id2
	val app1 = NT.mk_nuprl_apply_term term nid1
	val app2 = NT.mk_nuprl_apply_term app1 nid2
	val lam1 = NT.mk_nuprl_lambda_term id2 app2
	val lam2 = NT.mk_nuprl_lambda_term id1 lam1
    in lam2
    end

fun mk_nuprl_bag_deq_term deq =
    let val set  = get_all_ids_list [deq]
	val id1  = A.newIdIdSet set "x"
	val id2  = A.newIdIdSet set "y"
	val nid1 = NT.mk_nuprl_variable_term id1
	val nid2 = NT.mk_nuprl_variable_term id2
	val bdeq = NT.mk_nuprl_simple_term "bag-eq"  [deq, nid1, nid2]
	val lam1 = NT.mk_nuprl_lambda_term id2 bdeq
	val lam2 = NT.mk_nuprl_lambda_term id1 lam1
    in lam2
    end


(* ------ EXPORT E# TO NUPRL ------ *)

val minStr = "this term should have been transformed"

(* We have this environment as a reference to avoid passing it around in ToNuprl. *)
val unifenv = ref (EN.initSolved ())
fun setUnifEnv x  = unifenv := x
fun getUnifEnv () = !unifenv

(* This map is used to record how many times an identifier has been declared in
 * a piece of code.  If we're converting a declaration and we know that id has
 * been declared twice then we know that to get its binding stored in the
 * unifenv, we need to look at the third binding of id from the start. *)
structure NBID = BinaryMapFn(type ord_key = string val compare = String.compare)

(* map identifiers to the number of times they've been defined *)
type map_nbids = int NBID.map

val nbId = ref (NBID.empty : map_nbids)
fun addOneId id =
    case NBID.find (!nbId, id) of
	NONE   => nbId := NBID.insert (!nbId, id, 1)
      | SOME n => nbId := NBID.insert (!nbId, id, n + 1)
fun getNbOfId id =
    case NBID.find (!nbId, id) of
	NONE => 0
      | SOME n => n
fun resetNbId () = nbId := NBID.empty

(* This map is used to record the parameters used in the declarared function.
 * The number of parameters used in a function is then used to define the arity
 * of the function. *)
structure CONS = BinaryMapFn(type ord_key = string val compare = String.compare)

(* userFP stands for 'user functions with their parameters' *)
val userFP = ref (CONS.empty : string list CONS.map)
fun addUserFP id params = userFP := CONS.insert (!userFP, id, params)
fun getUserFP isbound id =
    if isbound id
    then NONE
    else CONS.find (!userFP, id)
fun resetUserFP () = userFP := CONS.empty

(* This set is used to record the parameters added to the declarations' bodies *)
structure DECFP = BinarySetFn(type ord_key = string val compare = String.compare)

val decFP = ref DECFP.empty
fun addDecFP param = decFP := DECFP.add (!decFP, param)
fun addDecFPList params = decFP := DECFP.addList (!decFP, params)
fun getDecFP () = DECFP.listItems (!decFP)
fun resetDecFP () = decFP := DECFP.empty

(*
(* This set is used to record the type parameters used in declarations' types. *)
structure PTYPES = BinarySetFn(type ord_key = string val compare = String.compare)

val ptypes = ref PTYPES.empty
fun addPType tparam = ptypes := PTYPES.add (!ptypes, tparam)
fun getPTypes () = PTYPES.listItems (!ptypes)
fun resetPTypes () = ptypes := PTYPES.empty
*)

(*
(* This set is used to record type parameters. *)
structure TPARAMS = BinarySetFn(type ord_key = string val compare = String.compare)

val tparams = ref TPARAMS.empty
fun addTParam tparam = tparams := TPARAMS.add (!tparams, tparam)
fun isTParam  tparam = TPARAMS.member (!tparams, tparam)
fun resetTParams  () = tparams := TPARAMS.empty
*)

(* This list is used to record all the headers as nuprl terms.
 * We do that to generate the full list of headers of an E# specification. *)
val headers = ref ([] : (string * NT.nuprl_term) list)
fun addHeader p = headers := p :: !headers
fun getHeaders () = !headers
fun resetHeaders () = headers := []

(* Object id of the esharp file within Nuprl. *)
val obidref = ref ""
fun setObid obid = obidref := obid
fun getObid () = !obidref
fun resetObid () = obidref := ""

(* Prefix to use for the generated Nuprl objects. *)
val prefref = ref ""
fun setPrefix pref = prefref := pref
fun getPrefix () = !prefref
fun resetPrefix () = prefref := ""

fun get_user_name opid =
    let val pref = getPrefix ()
    in if pref = ""
       then opid
       else pref ^ "_" ^ opid
    end

fun mk_nuprl_user_term_lvlop opid terms lvlop =
    let val obid    = getObid ()
	val name    = get_user_name opid
	val bterms  = map (fn t => ([], t)) terms
	val params1 = if obid = "" then [] else [NT.mk_nuprl_obid_parameter obid]
	val params2 = case lvlop of SOME i => [NT.mk_nuprl_level_exp_parameter i] | NONE => []
    in  NT.mk_nuprl_term (name, params1 @ params2) bterms
    end

fun mk_nuprl_user_term opid terms = mk_nuprl_user_term_lvlop opid terms NONE

fun reset () =
    (resetNbId    ();
     resetUserFP  ();
     resetDecFP   ();
     resetHeaders ();
     resetObid    ();
     resetPrefix  ();
     ())


(* ------ BEGIN CONTEXT ------ *)

structure IDENTS = BinarySetFn(type ord_key = string val compare = String.compare)
structure MAPID  = BinaryMapFn(type ord_key = string val compare = String.compare)

(* map identifier to their programs *)
structure PROGS = MAPID
type map_prog = NT.nuprl_term PROGS.map

(* set of bounds identifier *)
structure BOUNDS = IDENTS
type set_bounds = BOUNDS.set

(* This map is used to record the declared equality deciders *)
structure EQDEC = MAPID
type map_eqdec = string EQDEC.map

(* This map is used to record the arity of imported constants. *)
structure ARITY = MAPID
type map_arity = (string * (string * string list) list) ARITY.map

(* This map is used to record the defined parameters along with their types *)
structure PARAMS = MAPID
type map_params = E.scheme PARAMS.map

(* This map is used to record the datatypes and their parameters *)
structure DATA = MAPID
type map_data = string list DATA.map

type context =
     {progs   : map_prog,
      nbids   : map_nbids,
      arity   : map_arity,
      params  : map_params,
      eqdec   : map_eqdec,
      bounds  : set_bounds,
      data    : map_data,
      ndefs   : NT.lib option}

val mk_empty_context : context =
    {progs   = PROGS.empty,
     nbids   = NBID.empty,
     arity   = ARITY.empty,
     params  = PARAMS.empty,
     eqdec   = EQDEC.empty,
     bounds  = BOUNDS.empty,
     data    = DATA.empty,
     ndefs   = NONE}

fun mk_context progs nbids arity params eqdec bounds data ndefs =
    {progs   = progs,
     nbids   = nbids,
     arity   = arity,
     params  = params,
     eqdec   = eqdec,
     bounds  = bounds,
     data    = data,
     ndefs   = ndefs}

fun add_prog_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} id prog =
    mk_context (PROGS.insert (progs, id, prog)) nbids arity params eqdec bounds data ndefs

fun add_progs_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} progs' =
    mk_context (PROGS.unionWith (fn (_, x) => x) (progs, progs')) nbids arity params eqdec bounds data ndefs

fun add_new_id_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} id =
    case NBID.find (nbids, id) of
	NONE   => mk_context progs (NBID.insert (nbids, id, 1))     arity params eqdec bounds data ndefs
      | SOME n => mk_context progs (NBID.insert (nbids, id, n + 1)) arity params eqdec bounds data ndefs

fun add_arity_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} id name ar =
    mk_context progs nbids (ARITY.insert (arity, id, (name, ar))) params eqdec bounds data ndefs

fun add_param_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} param typ =
    mk_context progs nbids arity (PARAMS.insert (params, param, typ)) eqdec bounds data ndefs

fun add_eqdec_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} tycon eq =
    mk_context progs nbids arity params (EQDEC.insert (eqdec, tycon, eq)) bounds data ndefs

fun rm_eqdec_context (context as {progs, nbids, arity, params, eqdec, bounds, data, ndefs}) tycon =
    (mk_context progs nbids arity params (#1 (EQDEC.remove (eqdec, tycon))) bounds data ndefs)
    handle _ => context

fun add_alldefs_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} alldefsop =
    mk_context progs nbids arity params eqdec bounds data alldefsop

fun get_alldefs_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} = ndefs

fun add_bound_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} id =
    mk_context progs nbids arity params eqdec (BOUNDS.add (bounds, id)) data ndefs

fun add_bounds_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} ids =
    mk_context progs nbids arity params eqdec (BOUNDS.addList (bounds, ids)) data ndefs

fun add_data_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} tycon prms =
    mk_context progs nbids arity params eqdec bounds (DATA.insert (data, tycon, prms)) ndefs

fun is_bound_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} id =
    BOUNDS.member (bounds, id)

fun is_data_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} id =
    DATA.find (data, id)

fun get_eqdec_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} tycon =
    EQDEC.find (eqdec, tycon)

fun get_arity_context (context as {progs, nbids, arity, params, eqdec, bounds, data, ndefs}) id =
    if is_bound_context context id
    then NONE
    else ARITY.find (arity, id)

fun get_param_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} param =
    PARAMS.find (params, param)

fun get_all_params_context {progs, nbids, arity, params, eqdec, bounds, data, ndefs} =
    PARAMS.foldri (fn (i, _, list) => i :: list) [] params

fun is_param_context context param =
    Option.isSome (get_param_context context param)

(* ------ END CONTEXT ------ *)


fun getArityArgs (A.N {kind = (A.ARGS, A.ARGS_EM),   label, value, regions, parents, children = []}) = []
  | getArityArgs (A.N {kind = (A.ARGS, A.ARGS_PSEQ), label, value, regions, parents, children})      =
    map (fn (A.N {kind = (A.ARG, A.ARG_A), label, value, regions, parents, children = [id, args, ty]}) =>
	    (case args of
		 A.N {kind = (A.ARGS, A.ARGS_EM), label, value, regions, parents, children = []} => (A.getIdIdent id, [])
	       | A.N {kind = (A.ARGS, A.ARGS_LSEQ), label, value, regions, parents, children}    =>
		 let val lst = map (fn (A.N {kind = (A.ARG, A.ARG_A), label, value, regions, parents, children = [id, args, ty]}) => A.getIdIdent id
				     | _ => raise EH.Impossible "unexpected term")
				   children
		 in (A.getIdIdent id, lst)
		 end
	       | _ => raise EH.Impossible "unexpected term")
	  | _ => raise EH.Impossible "unexpected term")
	children
  | getArityArgs _ = raise EH.Impossible "unexpected term"

fun getArityTyvarSeq (A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_EM),  label, value, regions, parents, children}) = 0
  | getArityTyvarSeq (A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_ONE), label, value, regions, parents, children}) = 1
  | getArityTyvarSeq (A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_SEQ), label, value, regions, parents, children}) = List.length children
  | getArityTyvarSeq term = raise EH.Impossible "unexpected term"

fun flattenExpApp (A.N {kind = (A.EXP, A.EXP_APP), label, value, regions, parents, children = [exp, atexp]}) =
    let val (term, terms) = flattenExpApp exp
    in (term, terms @ [atexp])
    end
  | flattenExpApp term = (term, [])

fun toNuprlListAtoms [] = NT.mk_nuprl_nil_term
  | toNuprlListAtoms (x :: xs) =
    let val t = NT.mk_nuprl_token_term x
    in NT.mk_nuprl_cons_term t (toNuprlListAtoms xs)
    end

fun toAtomList atoms =
    case A.getKind atoms of
	(A.ATOMS, A.ATOMS_ATOMS) => toNuprlListAtoms (A.splitAtoms (A.getValue atoms))
      | _ => raise EH.Impossible ""

fun getIdPat n pat =
    case A.getIdPat pat of
	SOME id => A.getIdIdent id
      | NONE => (print (Int.toString n ^ "\n" ^ A.toString pat); raise EH.Impossible "")

(* TODO: we don't need this function, we can just gather the free params, while
 * converting the E# code, because we anyway make sure that the bound variables
 * don't clash with the parameter names. *)
fun getFreeParams context binders (A.N {kind = (A.ID, A.ID_VID), label, value, regions, parents, children = []})=
    if not (is_param_context context value)
       orelse
       List.exists (fn x => x = value) binders
    then []
    else [value]
  | getFreeParams context binders (A.N {kind = (A.EXP, A.EXP_LET), label, value, regions, parents, children = [bind, exp]}) =
    (case bind of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e]} =>
	 if A.termIsEmParam a
	 then let val fid = A.getIdIdent f
		  val params1 = getFreeParams context binders e
		  val params2 = getFreeParams context (fid :: binders) exp
	      in params1 @ params2
	      end
	 else raise EH.Impossible minStr
       | A.N {kind = (A.BIND, A.BIND_PAT), label, value, regions, parents, children = [p, e]} =>
	 let val idents =
		 case p of
		     A.N {kind = (A.PAT, A.PAT_ATPAT), label, value, regions, parents, children = [atpat]} =>
		     (case atpat of
			  A.N {kind = (A.ATPAT, A.ATPAT_TUPLE), label, value, regions, parents, children} => map (getIdPat 1) children
			| A.N {kind = (A.ATPAT, A.ATPAT_ID), label, value, regions, parents, children = [id]} => [A.getIdIdent id]
			| term => (print (A.toString term ^ "\n"); raise EH.Impossible minStr))
		   | _ => raise EH.Impossible minStr
	     val params1 = getFreeParams context binders e
	     val params2 = getFreeParams context (idents @ binders) exp
	 in params1 @ params2
	 end
       | _ => raise EH.Impossible minStr)
  | getFreeParams context binders (A.N {kind = (A.EXP, A.EXP_CLASS), label, value, regions, parents, children = [bind, exp]}) = raise EH.Impossible minStr
  | getFreeParams context binders (A.N {kind = (A.EXP, A.EXP_LETR), label, value, regions, parents, children = [bind, exp]}) =
    (case bind of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e]} =>
	 if A.termIsEmParam a
	 then let val fid = A.getIdIdent f
		  val params1 = getFreeParams context (fid :: binders) exp
		  val params2 = getFreeParams context (fid :: binders) e
	      in params1 @ params2
	      end
	 else raise EH.Impossible minStr
       | _ => raise EH.Impossible minStr)
  | getFreeParams context binders  (A.N {kind = (A.EXP, A.EXP_LAMBDA), label, value, regions, parents, children = [pat, exp]}) =
    getFreeParams context ((getIdPat 2 pat) :: binders) exp
  | getFreeParams context binders (A.N {kind = (A.EXP, A.EXP_BINDING), label, value, regions, parents, children = [exp1, pat, exp2]}) =
    let val params1 = getFreeParams context binders exp1
	val params2 = getFreeParams context ((getIdPat 3 pat) :: binders) exp2
    in params1 @ params2
    end
  | getFreeParams context binders (A.N {kind, label, value, regions, parents, children}) = List.concat (map (getFreeParams context binders) children)

structure BTV = BinaryMapFn(type ord_key = E.ityvar val compare = Int.compare)

fun toNuprlIty i force context vars (E.ITYVAR (ityvar, eq)) ityvarset etyvarset =
    (case BTV.find (ityvarset, ityvar) of
	 SOME str => NT.mk_nuprl_variable_term str
       | NONE => if force
		 then NT.mk_nuprl_variable_term (Int.toString ityvar)
		 else (print ("("  ^ Int.toString i ^ ")\n"); raise EH.Impossible "type variables are not allowed to be converted to nuprl"))
  (* We can have a type variable because we have a genuine generality or the
   * ityvar can come from a mono bind.  We might have to do more that recording
   * the type variable associated to a label.
   * We should be able to handle:
   *
   *     MSGS (``foo` : 'a List) ;;
   *     let m = MSG(``foo``, []) ;;
   *
   * But we don't. *)
  | toNuprlIty i force context vars (E.ITYETV (etyvar, lab)) ityvarset etyvarset =
    if E.isinETVS (etyvarset, etyvar)
    then NT.mk_nuprl_variable_term etyvar
    else raise EH.Impossible "type variables are not allowed to be converted to nuprl "
  | toNuprlIty i force context vars (E.ITYCON (E.ITYSEQSEQ (itys, _), E.ITYCONNAM (tyconnam, _, _))) ityvarset etyvarset =
    let val n_itys = map (fn ity => toNuprlIty i force context vars ity ityvarset etyvarset) itys
    in case (tyconnam, n_itys) of
           ("*",     nt1 :: nt2 :: nts) => foldl (fn (nty, prd) => NT.mk_nuprl_prod_term prd nty) (NT.mk_nuprl_prod_term nt1 nt2) nts
	 | ("*",     []) => NT.mk_nuprl_unit_term
	 | ("->",    [nt1, nt2]) => NT.mk_nuprl_fun_term   nt1 nt2
	 | ("+",     [nt1, nt2]) => NT.mk_nuprl_union_term nt1 nt2
	 | ("List",  [nt]) => NT.mk_nuprl_list_term  nt
	 | ("Class", [nt]) => NT.mk_nuprl_class_term nt
	 | ("Bag",   [nt]) => NT.mk_nuprl_bag_term   nt
	 | ("Deq",   [nt]) => NT.mk_nuprl_deq_term   nt
	 | ("Int",   []) => NT.mk_nuprl_int_term
	 | ("Bool",  []) => NT.mk_nuprl_bool_term
	 | ("Real",  []) => NT.mk_nuprl_real_term
	 | ("Msg",   []) => NT.mk_nuprl_msg_term "i"
	 | ("Unit",  []) => NT.mk_nuprl_unit_term
	 | ("Atom",  []) => NT.mk_nuprl_atom_term
	 | ("Type",  []) => NT.mk_nuprl_type_term "i"
	 | ("Prop",  []) => NT.mk_nuprl_prop_term "i"
	 | ("Nat",   []) => NT.mk_nuprl_nat_term
	 | ("Loc",   []) => NT.mk_nuprl_loc_term
	 | (tycon, args) =>
	   (case is_data_context context tycon of
		SOME prms =>
		let val a_terms = map NT.mk_nuprl_variable_term prms
		    val _ = addDecFPList prms
		in if List.null args
		   then mk_nuprl_user_term tycon a_terms
		   else raise Fail "toNuprlIty"
		end
	      | NONE =>
		if IDENTS.member (vars, tycon)
		then if List.null args
		     then NT.mk_nuprl_variable_term tycon
		     else raise Fail "toNuprlIty"
		else if is_param_context context tycon
		then let val _ = addDecFP tycon
		     in foldl (fn (arg, term) => NT.mk_nuprl_apply_term term arg)
			      (NT.mk_nuprl_variable_term tycon)
			      args
		     end
		else NT.mk_nuprl_simple_term tycon args)
    end
  | toNuprlIty i force context vars (E.ITYDEP (ity, _)) ityvarset etyvarset = toNuprlIty i force context vars ity ityvarset etyvarset
  | toNuprlIty i force context vars ity _ _ = (print ("*** (" ^ Int.toString i ^ ") " ^ E.toStringIty ity ^ "\n"); raise EH.Impossible "the type should be fully built up, no variable")

fun toNuprlIty_em i force context typ =
    toNuprlIty i force context IDENTS.empty typ BTV.empty E.emptyETVS

(* transforms a type into an equality decider *)
fun toDeq context (E.ITYVAR (ityvar, eq)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl"
  | toDeq context (E.ITYETV (etyvar, lab)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl "
  | toDeq context (E.ITYCON (E.ITYSEQSEQ (itys, _), E.ITYCONNAM (tyconnam, _, _))) =
    let fun toIty ity = toNuprlIty_em 0 false context (E.stripIty ity)
	val n_deqs = map (fn ity => (toIty ity, toDeq context ity)) itys
    in case (tyconnam, n_deqs) of
	   ("*", []) => NT.mk_nuprl_unit_deq_term
	 | ("*",     x :: y :: deqs) =>
	   (case List.rev n_deqs of
		(nty1, deq1) :: (nty2, deq2) :: deqs =>
		let val nty0 = NT.mk_nuprl_prod_term nty2 nty1
		    val deq0 = NT.mk_nuprl_product_deq_term nty2 nty1 deq2 deq1
		    val (_, deq) =
			foldl (fn ((nty', deq'), (nty, deq)) =>
				  (NT.mk_nuprl_prod_term nty' nty,
				   NT.mk_nuprl_proddeq_term deq' deq))
			      (nty0, deq0)
			      deqs
		in deq
		end
	      | _ => raise EH.Impossible "product type should have at least 2 arguments")
	 | ("->",    [(nty1, deq1), (nty2, deq2)]) => raise EH.Impossible "functional type does not have an equality decider"
	 | ("+",     [(nty1, deq1), (nty2, deq2)]) => NT.mk_nuprl_union_deq_term nty1 nty2 deq1 deq2
	 | ("Class", [(nty, deq)]) => raise EH.Impossible "Class does not have an equality decider"
	 | ("List",  [(nty, deq)]) => NT.mk_nuprl_list_deq_term deq
	 | ("Bag",   [(nty, deq)]) => mk_nuprl_bag_deq_term  deq
	 | ("Int",   []) => NT.mk_nuprl_int_deq_term
	 | ("Bool",  []) => NT.mk_nuprl_bool_deq_term
	 | ("Real",  []) => raise EH.Impossible "Real does not have an equality decider"
	 | ("Msg",   []) => raise EH.Impossible "Msg does not have an equality decider"
	 | ("Unit",  []) => NT.mk_nuprl_unit_deq_term
	 | ("Atom",  []) => NT.mk_nuprl_atom_deq_term
	 | ("Type",  []) => raise EH.Impossible "Type does not have an equality decider"
	 | ("Nat",   []) => NT.mk_nuprl_nat_deq_term
	 | ("Loc",   []) => NT.mk_nuprl_loc_deq_term
	 | (_, _) =>
	   (case get_eqdec_context context tyconnam of
		NONE =>
		let val str = "no equality decider for " ^ tyconnam
		    val _   = print (str ^ "\n")
		in raise EH.Impossible str
		end
	      | SOME deq =>
		if is_param_context context deq
		then let val _ = addDecFPList [deq]
		     in NT.mk_nuprl_applies_term (NT.mk_nuprl_variable_term deq) (map (fn (_, x) => x) n_deqs)
		     end
		else NT.mk_nuprl_applies_term (NT.mk_nuprl_simple_term deq [])  (map (fn (_, x) => x) n_deqs))
    end
  | toDeq context ity = (print ("*** " ^ E.toStringIty ity ^ "\n"); raise EH.Impossible "the type should be fully built up, no variable")

(*
(* transforms a type and two terms into a boolean. *)
fun toEqDeq nt1 nt2 (E.ITYVAR (ityvar, eq)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl"
  | toEqDeq nt1 nt2 (E.ITYETV (etyvar, lab)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl "
  | toEqDeq nt1 nt2 (E.ITYCON (E.ITYSEQSEQ (itys, _), E.ITYCONNAM (tyconnam, _, _))) =
    let val n_deqs = map (fn ity => toDeq ity) itys
    in case (tyconnam, n_deqs) of
           ("*",     [deq1, deq2]) => NT.mk_nuprl_eq_prod_term deq1 deq2 nt1 nt2
	 | ("->",    [deq1, deq2]) => raise EH.Impossible "-> does not have an equality decider"
	 | ("+",     [deq1, deq2]) => NT.mk_nuprl_eq_union_term deq1 deq2 nt1 nt2
	 | ("Class", [deq]) => raise EH.Impossible "Class does not have an equality decider"
	 | ("List",  [deq]) => NT.mk_nuprl_eq_list_term deq nt1 nt2
	 | ("Bag",   [deq]) => NT.mk_nuprl_eq_bag_term deq nt1 nt2
	 | ("Int",   []) => NT.mk_nuprl_eq_int_term nt1 nt2
	 | ("Bool",  []) => NT.mk_nuprl_eq_bool_term nt1 nt2
	 | ("Real",  []) => raise EH.Impossible "Real does not have an equality decider"
	 | ("Msg",   []) => raise EH.Impossible "Msg does not have an equality decider"
	 | ("Unit",  []) => raise EH.Unimplemented ""
	 | ("Atom",  []) => NT.mk_nuprl_eq_atom_term nt1 nt2
	 | ("Type",  []) => raise EH.Impossible "Type does not have an equality decider"
	 | ("Nat",   []) => NT.mk_nuprl_eq_int_term nt1 nt2
	 | ("Loc",   []) => NT.mk_nuprl_eq_loc_term nt1 nt2
	 | (_, _) =>
	   (case getEqDec tyconnam of
		NONE => raise EH.Impossible ("no equality decider for " ^ tyconnam)
	      | SOME deq =>
		let val ndeq = NT.mk_nuprl_simple_term deq n_deqs
		    val app1 = NT.mk_nuprl_apply_term ndeq nt1
		    val app2 = NT.mk_nuprl_apply_term app1 nt2
		in app2
		end)
    end
  | toEqDeq nt1 nt2 ity = (print ("*** " ^ E.toStringIty ity ^ "\n"); raise EH.Impossible "the type should be fully built up, no variable")
*)

datatype alpha = NEXT of string * (unit -> alpha)

val lstAlpha =
    ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
     "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z"]

fun getNewTyVarName [] suff () = getNewTyVarName lstAlpha (suff ^ "'") ()
  | getNewTyVarName (x :: xs) suff () =
    let val f = getNewTyVarName xs suff
    in NEXT (x ^ suff, f)
    end

val getNewTyVarName = getNewTyVarName lstAlpha ""

(* NOTE: idorset should be empty, we shouldn't have any or types in the types
 * of the top declarations. *)
fun toNuprlScheme context (E.VPOLY ((ityvarset, etyvarset, idorset), (overseq, ity))) =
    let val ity'  = E.stripIty ity
	val (nts1, ren) =
	    let val (nts, ren, _) =
		    E.foldrITVS (fn (tv, (nts, ren, f)) =>
				    let val NEXT (str, f') = f ()
				    in (str :: nts, BTV.insert (ren, tv, str), f')
				    end)
				([], BTV.empty, getNewTyVarName)
				ityvarset
	    in (nts, ren)
	    end
	val nts2  = E.listItemsETVS etyvarset
	val n_ity = toNuprlIty 1 false context IDENTS.empty ity' ren etyvarset
    in (n_ity, nts2 @ nts1)
    end
  | toNuprlScheme context (E.VODEC _) = raise EH.Impossible "overloaded schemes are not allowed to be converted into NuPrl"
  | toNuprlScheme context (E.VMONO _) = raise EH.Impossible "monomorphic schemes are not allowed to be converted into NuPrl"

(* NOTE: ityvarset and idorset should be empty. *)
fun schemeParamToHead context (E.VPOLY ((ityvarset, etyvarset, idorset), (overseq, ity))) =
    let val ity'  = E.stripIty ity
	val nts   = E.listItemsETVS etyvarset
	val n_ity = toNuprlIty 2 false context IDENTS.empty ity' BTV.empty etyvarset
    in (n_ity, nts)
    end
  | schemeParamToHead context (E.VODEC _) = raise EH.Impossible "overloaded schemes are not allowed to be converted into NuPrl"
  | schemeParamToHead context (E.VMONO _) = raise EH.Impossible "monomorphic schemes are not allowed to be converted into NuPrl"

structure SETVARS = BinarySetFn(type ord_key = string val compare = String.compare)

fun getVariablesNuprlType term =
    if NT.is_nuprl_variable_term term
    then SETVARS.singleton (NT.dest_variable term)
    else let val (opid, sterms) = NT.dest_simple_null_term term
	 in foldr (fn (term, vars) =>
		      SETVARS.union (getVariablesNuprlType term, vars))
		  SETVARS.empty
		  sterms
	 end

fun findNlpsInType term typ (args, set) =
    let val opid = NT.opid_of_term typ
    in if opid = "eclass"
       then let val (_, _, _, Y) = NT.dest_eclass typ
		val nlp = NT.mk_nuprl_nlp_term Y term
		val prg = NT.mk_nuprl_programmable_term Y term
	    in ([(nlp, prg)], args, set)
	    end
       else if opid = "product"
       then let val (left, right) = NT.dest_product typ
		val (nlps1, args1, set1) = findNlpsInType (NT.mk_nuprl_pi1_term term)
							  left
							  ([], set)
		val (nlps2, args2, set2) = findNlpsInType (NT.mk_nuprl_pi2_term term)
							  right
							  ([], set1)
		val nlps  = nlps1 @ nlps2
		val args' = if null nlps then [] else args
	    in (nlps, args' @ args1 @ args2, set2)
	    end
       else if opid = "function"
       then let val (left, right) = NT.dest_function typ
		val id    = A.newId ()
		val v     = NT.mk_nuprl_variable_term id
		val args' = (id, left) :: args
		val set'  = IDS.add (set, id)
	    in findNlpsInType (NT.mk_nuprl_apply_term term v) right (args', set')
	    end
       else ([], [], set)
    end

fun genNormLocProg set class nity args =
    let val opid = NT.opid_of_term nity
    in if opid = "eclass"
       then let val (msg, es, e, X) = NT.dest_eclass nity
		val v_args   = map (fn (v, typ) => NT.mk_nuprl_variable_term v) args
		val applies  = NT.mk_nuprl_applies_term class v_args
		val setvars  = getVariablesNuprlType X
		val vars     = SETVARS.listItems setvars
		(*val valueall = NT.mk_nuprl_valueall_type_term X*)
		val nlp      = NT.mk_nuprl_nlp_term X applies
		val prg      = NT.mk_nuprl_programmable_term X applies
		(*val vallimp  = NT.mk_nuprl_implies_term valueall nlp*)
		val (nlpargs, oargs, _)  =
		    foldr (fn ((v, typ), (nlps, args, set)) =>
			      let val term = NT.mk_nuprl_variable_term v
				  val (nlps', args', set') = findNlpsInType term typ ([], set)
			      in (nlps' @ nlps, args' @ args, set')
			      end)
			  ([], [], set)
			  args
		val (nlpsimp, prgsimp)  =
		    foldr (fn ((leftN, leftP), (rightN, rightP)) =>
			      (NT.mk_nuprl_implies_term leftN rightN,
			       NT.mk_nuprl_implies_term leftP rightP))
			  (nlp, prg) (*vallimp*)
			  nlpargs
		val (nlp_term, prg_term) =
		    foldl (fn ((v, typ), (termN, termP)) =>
			      (NT.mk_nuprl_all_term typ (v, termN),
			       NT.mk_nuprl_all_term typ (v, termP)))
			  (nlpsimp, prgsimp)
			  (oargs @ args)
	    in SOME (vars, nlp_term, prg_term)
	    end
       else if opid = "function"
       then let val (t1, t2) = NT.dest_function nity
		val narg  = "x" ^ Int.toString (List.length args)
		val id    = A.newIdIdSet set narg
		val args' = args @ [(id, t1)]
		val set'  = IDS.add (set, id)
	    in genNormLocProg set' class t2 args'
	    end
       else NONE
    end

fun genAbstractionAndWF i context value id n_exp exp schemeop premises programmable data lvlop =
    let val eid    = if A.isInternalId value then value else id
	(*val _ = print ("gen-abs (" ^ Int.toString i ^ ")\n")*)
	val ((n_ity, bounds), context') =
	    let val nb = getNbOfId eid
	    in case schemeop of
		   SOME scheme => (scheme, context)
		 | NONE =>
		   (case EN.getScheme eid (nb + 1) (getUnifEnv ()) of
			NONE =>
			(print (":" ^ Int.toString i ^ ":" ^ value ^ ":" ^ id ^ ":" ^ eid);
			 raise EH.Impossible ("Couldn't find " ^ eid ^ "'s environment"))
		      | SOME scheme =>
			let val _ = addOneId eid
			in (toNuprlScheme context scheme,
			    add_new_id_context context eid)
			end)
	    end
	val _      = addDecFPList (getFreeParams context [] exp)
	val parms  = getDecFP ()
	val n_prs  = map NT.mk_nuprl_variable_term parms
	val n_id   = mk_nuprl_user_term_lvlop id n_prs lvlop
	val n_abs  = NT.mk_nuprl_iabstraction_term n_id n_exp
	val n_typ  = NT.mk_nuprl_type_term "i'"
	val n_atyp = NT.mk_nuprl_valuealltype_term "i"
	val n_btyp = NT.mk_nuprl_valuealltype_term "i'"
	val nprms  = map (fn param =>
			     case get_param_context context param of
				 SOME scheme => (param, schemeParamToHead context scheme)
			       | NONE => raise EH.Impossible "")
			 parms
	fun toAllPrmsWf term =
	    foldr (fn ((param, (nity, nts)), n_term) =>
		      foldr (NT.mk_nuprl_uall_term n_typ)
			    (let val typ =
				      if NT.is_nuprl_type_term nity
				      then n_atyp
				      else nity
			      in NT.mk_nuprl_uall_term typ (param, n_term)
			      end (*NT.mk_nuprl_uall_term nity (param, n_term)*))
			    nts)
		  term
		  nprms
	fun toAllPrmsNlp term =
	    foldr (fn ((param, (nity, nts)), n_term) =>
		      ((*print ("[" ^ param ^ "]\n");
		       print (toStringTerm nity ^ "\n");*)
		       foldr (NT.mk_nuprl_all_term n_typ)
			     (let val typ =
				      if NT.is_nuprl_type_term nity
				      then n_atyp
				      else nity
			      in NT.mk_nuprl_all_term typ (param, n_term)
			      end)
			     nts))
		  term
		  nprms
	(*val n_thm  = let val n_mem   = NT.mk_nuprl_member_term n_ity n_id
			 val all     = foldr (NT.mk_nuprl_uall_term n_typ)
					     n_mem
					     bounds
			 val allprms = toAllPrmsWf all
			 val wfname  = get_user_name id ^ "_wf"
		     in NT.mk_nuprl_itheorem_term wfname allprms
		     end*)
	val (n_thm, set) =
	    let  val set = get_all_ids_list [n_id, n_ity]
		 val (term, n_ity', alls, asmps, set') =
		    foldl (fn (premise, (term, n_ity, alls, asmps, set)) =>
			      if NT.is_nuprl_function_term n_ity
			      then let val (A, B) = NT.dest_function n_ity
				       val v     = A.newIdIdSet set "v"
				       val tv    = NT.mk_nuprl_variable_term v
				       val term' = NT.mk_nuprl_apply_term term tv
				   in (term', B, alls @ [(v, A)], asmps @ [premise tv], IDS.add (set, v))
				   end
			      else raise Fail "genabstraction:too-many-premises")
			  (n_id, n_ity, [], [], set)
			  premises
		val isect   = foldr (NT.mk_nuprl_isect_term n_typ)
				    n_ity'
				    bounds
		val n_mem   = NT.mk_nuprl_member_term isect term
		val n_mem1  = foldr (fn (prem, term) => NT.mk_nuprl_uimplies_term prem term) n_mem asmps
		val n_mem2  = foldr (fn ((v, typ), term) => NT.mk_nuprl_uall_term typ (v, term)) n_mem1 alls
		val allprms = toAllPrmsWf n_mem2
		val wfname  = get_user_name id ^ "_wf"
	    in (NT.mk_nuprl_itheorem_term wfname allprms, set')
	    end
	val n_nrms = case genNormLocProg set n_id n_ity [] of
			 SOME (vars, nlp_term, prg_term) =>
			 let val (n_all_N, n_all_P) =
				 foldr (fn (v, (tN, tP)) =>
					   if List.exists (fn x => x = v) vars
					   then (NT.mk_nuprl_all_term n_btyp (v, tN),
						 NT.mk_nuprl_all_term n_btyp (v, tP))
					   else (NT.mk_nuprl_all_term n_typ  (v, tN),
						 NT.mk_nuprl_all_term n_typ  (v, tP)))
				       (nlp_term, prg_term)
				       bounds
			     val all_nlp = toAllPrmsNlp n_all_N
			     val all_prg = toAllPrmsNlp n_all_P
			     val nlp_lem = NT.mk_nuprl_itheorem_term (get_user_name id ^ "_nlp") all_nlp
			     val prop    = NT.mk_nuprl_iproperty_term "extract_required" (NT.mk_nuprl_ibool_term true)
			     val cons    = NT.mk_nuprl_icons_cons_term prop NT.mk_nuprl_icons_nil_term
			     val ipropsn = NT.mk_nuprl_iinclude_properties_term cons nlp_lem
			     val prg_lem = NT.mk_nuprl_itheorem_term (get_user_name id ^ "_programmable") all_prg
			     val ipropsp = NT.mk_nuprl_iinclude_properties_term cons prg_lem
			 in if programmable
			    then [ipropsn, ipropsp]
			    else [ipropsn]
			 end
		       | NONE => []
	val context' = if data
		       then add_data_context context id parms
		       else let val _ = addUserFP id parms
			    in context
			    end
    in (n_abs, n_thm, n_nrms, context')
    end

fun genAbstractionAndWF' i context value id n_exp exp premises =
    let val (abs, thm, rest, _) = genAbstractionAndWF i context value id n_exp exp NONE premises false false NONE
    in (abs, thm, rest)
    end

val dummy_dot_term = A.mk_new_dum_term A.DOTS_D "" [] []

fun genAbstractionAndWFDots i context value id n_exp premises =
    genAbstractionAndWF' i context value id n_exp dummy_dot_term premises

fun genInvariant context name cls (nargs, args) (nprop, prop) T =
    let val _      = app (fn x => addDecFPList (getFreeParams context [] x)) args
	val _      = addDecFPList (getFreeParams context [] prop)
	val parms  = getDecFP ()
	val ncls   = foldl (fn (arg, cls) => NT.mk_nuprl_apply_term cls arg)
			   (NT.mk_nuprl_variable_term cls)
			   nargs
	val tparms = map NT.mk_nuprl_variable_term parms
	val n_typ  = NT.mk_nuprl_type_term "i'"
	val n_atyp = NT.mk_nuprl_valuealltype_term "i"
	val nprms  = map (fn param =>
			     case get_param_context context param of
				 SOME scheme => (param, schemeParamToHead context scheme)
			       | NONE => raise EH.Impossible "")
			 parms
	val set    = get_all_ids_list [ncls, nprop, T]
	val v      = A.newIdIdSet set "v"
	val eo     = A.newIdIdSet set "eo"
	val e      = A.newIdIdSet set "e"
	val tv     = NT.mk_nuprl_variable_term v
	val teo    = NT.mk_nuprl_variable_term eo
	val te     = NT.mk_nuprl_variable_term e
	val clsrel = NT.mk_nuprl_classrel_term teo T ncls te tv
	val prp    = NT.mk_nuprl_apply_term nprop tv
	val imp    = NT.mk_nuprl_implies_term clsrel nprop
	val all    = NT.mk_nuprl_all_term T (v, imp)
	val term1  =
	    foldr (fn ((param, (nity, nts)), n_term) =>
		      let val typ = if NT.is_nuprl_type_term nity then n_atyp else nity
		      in foldr (NT.mk_nuprl_all_term n_typ)
			       (NT.mk_nuprl_all_term typ (param, n_term))
		      	       nts
		      end)
		  all
		  nprms
	val term2  = NT.mk_nuprl_all_term (NT.mk_nuprl_esE_term teo) (e, term1)
	val term3  = NT.mk_nuprl_all_term (NT.mk_nuprl_event_orderingp_term "i") (eo, term2)
	val thname = get_user_name name
	val thm    = NT.mk_nuprl_itheorem_term thname term3
    in thm
    end

fun toNuprlCombComp n_exp n_exps bprior optop =
    let val list =
	    foldr (fn ((exp, prg), list) =>
		      NT.mk_nuprl_cons_term exp list)
		  NT.mk_nuprl_nil_term
		  n_exps
	val set  = get_all_ids_list [n_exp, list]
	val id   = A.newIdIdSet set "n"
	val tid  = NT.mk_nuprl_variable_term id
	val sel  = NT.mk_nuprl_lambda_term id (NT.mk_nuprl_select_term tid list)
	(* NOTE: sel is the list of classes as a function. *)
	val fid  = A.newIdIdSet set "w"
	val sid  = A.newIdIdSet set "s"
	val tfid = NT.mk_nuprl_variable_term fid
	val tsid = NT.mk_nuprl_variable_term sid
	val lid  = A.newIdIdSet set "loc"
	val thd  = let val tlid = NT.mk_nuprl_variable_term lid
		   in NT.mk_nuprl_apply_term n_exp tlid
		   end
	val nt'  = #1 (foldr (fn (_, (nt, n)) =>
				 let val n_term = NT.mk_nuprl_natural_number_term n
				     val app1   = NT.mk_nuprl_apply_term tfid n_term
				     val app2   = NT.mk_nuprl_apply_term nt app1
				 in (app2, n + 1)
				 end)
			     (thd, 0)
			     n_exps)
	val app  = NT.mk_nuprl_apply_term nt' tsid
	val lam0 = NT.mk_nuprl_lambda_term sid app
	val lam1 = NT.mk_nuprl_lambda_term fid lam0
	val lam2 = NT.mk_nuprl_lambda_term lid lam1
    in if bprior
       then case optop of
		SOME (e_opt, p_opt) => NT.mk_nuprl_rec_comb_term lam2 sel e_opt
	      | NONE                => NT.mk_nuprl_rec_combined_loc_class_term lam2 sel
       else NT.mk_nuprl_combined_loc_class_term lam2 sel
    end

fun double x = (x, x)

fun toLift n exp blift bconcat bprior =
    let val (liftA, concatA, liftB, concatB) =
	    case n of
		0 => (NT.mk_nuprl_lifting_loc1_term,
		      NT.mk_nuprl_concat_lifting_loc1_term,
		      NT.mk_nuprl_lifting_loc0_term,
		      NT.mk_nuprl_concat_lifting_loc0_term)
	      | 1 => (NT.mk_nuprl_lifting_loc2_term,
		      NT.mk_nuprl_concat_lifting_loc2_term,
		      NT.mk_nuprl_lifting_loc1_term,
		      NT.mk_nuprl_concat_lifting_loc1_term)
	      | 2 => (NT.mk_nuprl_lifting_loc3_term,
		      NT.mk_nuprl_concat_lifting_loc3_term,
		      NT.mk_nuprl_lifting_loc2_term,
		      NT.mk_nuprl_concat_lifting_loc2_term)
	      | 3 => (NT.mk_nuprl_lifting_loc4_term,
		      NT.mk_nuprl_concat_lifting_loc4_term,
		      NT.mk_nuprl_lifting_loc3_term,
		      NT.mk_nuprl_concat_lifting_loc3_term)
	      | _ => let val nt =
			     if bprior
			     then NT.mk_nuprl_natural_number_term (n + 1)
			     else NT.mk_nuprl_natural_number_term n
		     in (NT.mk_nuprl_lifting_loc_gen_term nt,
			 NT.mk_nuprl_concat_lifting_loc_gen_term nt,
			 NT.mk_nuprl_lifting_loc_gen_term nt,
			 NT.mk_nuprl_concat_lifting_loc_gen_term nt)
		     end
	fun F flift fconcat =
	    if blift
	    then if bconcat
		 then fconcat exp
		 else flift exp
	    else exp
    in if bprior
       then F liftA concatA
       else F liftB concatB
    end

val default_strict = NT.mk_nuprl_inr_term NT.mk_nuprl_it_term

fun toRecCombProg Xprs F bprior opt typ =
    let val n     = List.length Xprs
	val set   = get_all_ids_list (F :: Xprs)
	val id    = A.newIdIdSet set "n"
	val tid   = NT.mk_nuprl_variable_term id
	val lst   = NT.mk_nuprl_finite_list_term Xprs
	val sel   = NT.mk_nuprl_select_term tid lst
	val FXprs = NT.mk_nuprl_lambda_term id sel
	(* FXprs is of the form: \n.[X1;X2;X3;X4][n] *)
	val num   = NT.mk_nuprl_integer_term n
	val l     = A.newIdIdSet set "l"
	val g     = A.newIdIdSet set "g"
	val p     = A.newIdIdSet set "p"
	val t_l   = NT.mk_nuprl_variable_term l
	val t_g   = NT.mk_nuprl_variable_term g
	val t_p   = NT.mk_nuprl_variable_term p
	val tl    = if bprior then [p]   else []
	val t_tl  = if bprior then [t_p] else []
	val tab   = List.tabulate (n, fn x => x)
	val apgs  = map (NT.mk_nuprl_apply_term t_g o NT.mk_nuprl_integer_term) tab
	(* apgs is of the form: [g 1;g 2; g 3;g 4] *)
	val app   = NT.mk_nuprl_applies_term F (t_l :: (apgs @ t_tl))
	(* app is of the form: (F l [g 1;g 2; g 3;g 4] p) *)
	val f     = NT.mk_nuprl_lambdas_term ([l, g] @ tl) app
	(* f is of the form: \l. \g. \p. (F l [g 1;g 2; g 3;g 4] p) *)
    in if bprior
       then case opt of
		SOME (_, init) => NT.mk_nuprl_rec_comb_prc_term typ num FXprs init f default_strict
	      | NONE => NT.mk_nuprl_rec_combined_loc_class1_prc_term typ num FXprs f
       else NT.mk_nuprl_loc_comb1_prc_term typ num FXprs f
    end

fun preToNuprlCombComp (n_exp, n_prg) exps bprior opt typ blift bconcat =
    let val (f0, f1, f2, f3) =
	    if bprior
	    then case opt of
		     SOME (e_opt, p_opt) =>
		     (NT.mk_nuprl_rec_combined0_loc_class_opt_term e_opt,
		      NT.mk_nuprl_rec_combined1_loc_class_opt_term e_opt,
		      NT.mk_nuprl_rec_combined2_loc_class_opt_term e_opt,
		      NT.mk_nuprl_rec_combined3_loc_class_opt_term e_opt)
		   | NONE => (NT.mk_nuprl_rec_combined0_loc_class_term,
			      NT.mk_nuprl_rec_combined1_loc_class_term,
			      NT.mk_nuprl_rec_combined2_loc_class_term,
			      NT.mk_nuprl_rec_combined3_loc_class_term)
	    else (NT.mk_nuprl_combined0_loc_class_term,
		  NT.mk_nuprl_combined1_loc_class_term,
		  NT.mk_nuprl_combined2_loc_class_term,
		  NT.mk_nuprl_combined3_loc_class_term)
	val n = List.length exps
	fun lift f = toLift n f blift bconcat bprior
	val class =
	    case exps of
		[] => f0 (lift n_exp)
	      | [(ne, _)] => f1 (lift n_exp) ne
	      | [(ne1, _), (ne2, _)] => f2 (lift n_exp) ne1 ne2
	      | [(ne1, _), (ne2, _), (ne3, _)] => f3 (lift n_exp) ne1 ne2 ne3
	      | _ => toNuprlCombComp (lift n_exp) exps bprior opt
	val prog = toRecCombProg (map (fn (_, x) => x) exps) (lift n_prg) bprior opt typ
    in (class, prog)
    end

fun get_pair_types [] map = []
  | get_pair_types ((term, etyp) :: lst) map =
    if NT.is_nuprl_variable_term term
    then let val var = NT.dest_variable term
	 in case MAPID.find (map, var) of
		SOME ntyp => (ntyp, etyp) :: (get_pair_types lst map)
	      | NONE => raise Fail ("get_pair_types:no-type-for-arg(" ^ var ^ ")")
	 end
    else raise Fail "get_pair_types:not-var"

fun set_missing var typ [] = []
  | set_missing var typ (lst as ((x : string, optyp) :: missings)) =
    if var = x
    then if Option.isSome optyp
	 then lst
	 else (x, SOME typ) :: missings
    else (x, optyp) :: (set_missing var typ missings)

fun get_type_vars [] missings = missings
  | get_type_vars ((ntyp, etyp) :: pairs) missings =
    if NT.is_nuprl_variable_term ntyp
    then let val var = NT.dest_variable ntyp
	 in get_type_vars pairs (set_missing var etyp missings)
	 end
    else if NT.is_nuprl_function_term ntyp
    then case E.destArrowType etyp of
	     SOME (eA, eB) =>
	     let val (nA, nB) = NT.dest_function ntyp
	     in get_type_vars ((nA, eA) :: (nB, eB) :: pairs) missings
	     end
	   | NONE => raise Fail "get_type_vars:types-do-not-match-function"
    else if NT.is_nuprl_bag_term ntyp
    then case E.destBagType etyp of
	     SOME eA =>
	     let val nA = NT.dest_bag ntyp
	     in get_type_vars ((nA, eA) :: pairs) missings
	     end
	   | NONE => raise Fail "get_type_vars:types-do-not-match-bag"
    else get_type_vars pairs missings

fun find_missings_args_tr opid (wf_lemma :: _) nsubs typ =
    (case NT.subterms wf_lemma of
	 [wf] =>
	 let val (alls, mem) = NT.dest_alls_ualls wf
	     val typ_map =
		 foldr
		     (fn ((var, typ, b), map) =>
			 if b (* it's a all and not a uall *)
			 then MAPID.insert (map, var, typ)
			 else map)
		     MAPID.empty
		     alls
	     val (prc_typ, prc) = NT.dest_member mem
	     val (prc_opid, prc_subs) = NT.dest_simple_term prc
	 in if prc_opid = opid ^ "-prc"
	    then let val n = List.length prc_subs - (nsubs + 1)
		 in if n >= 0
		    then let val (targs, tcls) = E.destArrowsType typ
			     val missings = List.take (prc_subs, n)
			     val terms = List.drop (prc_subs, n + 1)
			 in if List.length targs = List.length terms
			    then let val pairs = get_pair_types (ListPair.zip (terms, targs)) typ_map
				     val missings' = map (fn t => (NT.dest_variable t, NONE)) missings
				 in map (fn (v, SOME x) => (v, x)
					  | (v, NONE) => raise Fail ("find_missings_args_tr:(" ^ v ^ ")missing"))
					(get_type_vars pairs missings')
				 end
			    else raise Fail "find_missings_args_tr:eml-type-does-not-match"
			 end
		    else raise Fail "find_missings_args_tr:prc-does-not-have-enough-subterms"
		 end
	    else raise Fail "find_missings_args_tr:prc-name-does-not-match"
	 end
       | _ => raise Fail "find_missings_args_tr:more-than-one-subterms-in-wf")
  | find_missings_args_tr _ _ _ _ = raise Fail "find_missings_args_tr:no-wf"

(* Transformation of a class into a program *)
fun class2prog_tr context abs class typ Xtyp =
    let val opid = NT.opid_of_term class
	(*val _ = print (":" ^ opid ^ "\n" ^ E.ppIty' Xtyp ^ "\n-\n")*)
    in if opid = "rec-comb"
       then let val (Xs, F, init) = NT.dest_rec_comb class
		val num =
		    if NT.is_nuprl_lambda_term Xs
		    then let val (v,B) = NT.dest_lambda Xs
			 in if NT.is_nuprl_select_term B
			    then let val (n,lst) = NT.dest_select B
				 in List.length (NT.dest_list lst)
				 end
			    else raise Fail "class2prog_tr"
			 end
		    else raise Fail "class2prog_tr"
		val n = NT.mk_nuprl_integer_term num
		(*val _ = print ("---rc\n")*)
	    in NT.mk_nuprl_rec_comb_prc_term typ n Xs init F default_strict
	    end
       else if opid = "bind-class"
       then raise Fail "class2prog_tr:bind-class"
       else if opid = "parallel-class"
       then raise Fail "class2prog_tr:parallel-class"
       else (*if opid = "let"
       then let val (exp1, v, exp2) = NT.dest_let class
		val prg1 = class2prog_tr abs exp1 ...
		val prg2 = class2prog_tr abs exp2 typ
	    in NT.mk_nuprl_let_term prog1 (v, prg2)
	    end
       else*)
	   let val prc = opid ^ "-prc"
	   in case NT.MAP.find (abs, prc) of
		  SOME [{id, sign, obid, lhs, rhs, wfs}] =>
		  let val (_,subs) = NT.dest_simple_term class
		      val ns = #2 (sign : NT.sign)
		      val nsubs = List.length subs
		  in if List.all (fn x => x = 0) ns
		     then if List.length ns = nsubs + 1
			  (* NOTE: we could do something smarter here because
			   * class X is also mentioned in X-prc_wf *)
			  then NT.mk_nuprl_simple_term prc (typ :: subs)
			  else if List.length ns > nsubs + 1
			  (* NOTE: Here we should really do something else! *)
			  then let (*val _ = print ("--" ^ opid ^ "--\n")*)
				   val missings = find_missings_args_tr opid wfs nsubs Xtyp
				   val typs =
				       map (fn (x, t) =>
					       let (*val _ = print ("   :" ^ x ^ "\n")*)
						   (*val _ = print ("   :" ^ E.ppIty' t ^ "\n")*)
					       in toNuprlIty_em 19 true context t
					       end)
					   missings
			       in NT.mk_nuprl_simple_term prc (typs @ (typ :: subs))
			       end
			  (* let val n = List.length ns - (List.length subs + 1)
				 val terms = map (fn _ => NT.mk_nuprl_it_term) (List.tabulate (n, fn _ => ()))
			     in NT.mk_nuprl_simple_term prc (terms @ (typ :: subs))
			     end*)
			  else raise Fail "class2prog:prc-1:prc-does-not-have-enough-parameters"
		     else raise Fail "class2prog:prc-2:prc-with-binders"
		  end
		| _ => raise Fail ("class2prog:prc-3(" ^ prc ^ ")missing")
	   end
	   (*handle _ =>
		  let (*val _ = print (">" ^ NT.opid_of_term class ^ "\n")*)
		      val (_, _, _, lhs, rhs) = NT.find_sign abs class
		  in class2prog_tr abs (NT.unfold_ab NONE class lhs rhs) typ
		  end*)
    end

and class2prog context label class =
    case get_alldefs_context context of
	SOME {abs, tof} =>
	let val id  = NT.opid_of_term class
	    val n   = List.length (NT.bterms_of_term class)
	    (*val (x,xs) = NT.dest_applies class*)
	    (*val _ = print (":" ^ NT.opid_of_term x ^ ":" ^ id ^ ":" ^ Int.toString (List.length xs) ^ "\n")*)
	    val ity =
		case EN.getItyAtLab (getUnifEnv ()) label of
		    NONE => raise Fail ("We shoud have generated a type for identifier " ^ id)
		  | SOME ity => E.stripIty ity
	    val (lst, ity') = E.destArrowsType ity
	    val prog = (* if n <> length lst then we've got to add apps to the class
			* to get a class and then put back some lambdas *)
		if List.length lst = n
		then case E.destClassType ity' of
			 SOME cls =>
			 let (*val _ = print ("--" ^ id ^ "\n")*)
			     val typ = toNuprlIty_em 15 true context cls
			 in class2prog_tr context abs class typ ity
			 end
		       | NONE => class
		else class
	in prog
	end
      | NONE => class

fun transfun #" " = "\\ "
  | transfun x = Char.toString x

val transform = String.translate transfun

fun protect #"(" = "\\("
  | protect #")" = "\\)"
  | protect #"{" = "\\{"
  | protect #"}" = "\\}"
  | protect #"\\" = "\\\\"
  | protect #":" = "\\:"
  | protect #";" = "\\;"
  | protect #"," = "\\,"
  | protect x = Char.toString x

val protect = String.translate protect

(* Transform a term into a nuprl term *)
fun toNuprlTerm context (term as A.N {kind = (A.SCON, A.SCON_INT),    label, value, regions, parents, children = []}) =
    (case Int.fromString value of
	 NONE   => raise EH.Impossible A.wrongFormat
       | SOME n => double (NT.mk_nuprl_integer_term n))
  | toNuprlTerm context (term as A.N {kind = (A.SCON, A.SCON_REAL),   label, value, regions, parents, children = []}) = raise EH.Unimplemented ""
  | toNuprlTerm context (term as A.N {kind = (A.SCON, A.SCON_ATOM),   label, value, regions, parents, children = []}) = double (NT.mk_nuprl_token_term value)
  | toNuprlTerm context (term as A.N {kind = (A.SCON, A.SCON_ATOMS),  label, value, regions, parents, children = []}) = raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.SCON, A.SCON_STRING), label, value, regions, parents, children = []}) = raise EH.Impossible minStr

  (* ID *)
  | toNuprlTerm context (term as A.N {kind = (A.ID, A.ID_VID),    label, value, regions, parents, children = []}) =
    (case value of
	 "true"  => double (NT.mk_nuprl_btrue_term)
       | "false" => double (NT.mk_nuprl_bfalse_term)
       | "nil"   => double (NT.mk_nuprl_nil_term)
       | "="     =>
	 (case EN.getItyAtLab (getUnifEnv ()) label of
	      NONE => raise EH.Impossible "We shoud have generated a type for this equality"
	    | SOME ity =>
	      let val ity1 = E.stripIty ity
		  val ity2 = Option.valOf (E.destDeqType ity1)
		      handle _ => (print (E.toStringIty ity1);
				   raise EH.Impossible "")
	      in double (toDeq context ity2)
	      end)
       | _ =>
	 (case get_arity_context context value of
	      NONE =>
	      (* NONE means that the id (value) is not a constant,
	       * it is a user defined function *)
	      (case getUserFP (is_bound_context context) value of
		   (* We sometimes have to add parameters to the argument list
		    * of a used function, if it does use the parameters. *)
		   NONE => double (NT.mk_nuprl_variable_term value)
		 (* NONE means that the id (value) is not a top-level dec
		  * but a local one. *)
		 | SOME params =>
		   let val a_terms = map NT.mk_nuprl_variable_term params
		       val _ = addDecFPList params
		   in double (mk_nuprl_user_term value a_terms)
		   end)
	    | SOME (idname, lst) =>
	      if is_param_context context value
	      then if List.null lst
		   then double (NT.mk_nuprl_variable_term value)
		   else raise EH.Impossible ""
	      else let val name =
			   case EN.getNameAtLab (getUnifEnv ()) label of
			       NONE => idname
			     | SOME name => name
		       val vars =
			   List.map (fn (id, args) =>
					let val set = IDS.addList (IDS.singleton name, args)
					in (A.newIdIdSet set id, args)
					end)
				    lst
		       val s_terms =
			   map (fn (id, args) =>
				   (args,
				    foldl (fn (arg, term) =>
					      let val t_arg = NT.mk_nuprl_variable_term arg
					      in NT.mk_nuprl_apply_term term t_arg
					      end)
					  (NT.mk_nuprl_variable_term id)
					  args))
			       vars
		       val term = NT.mk_nuprl_term (name, []) s_terms
		   in double (foldr (fn ((id, _), term) =>
					NT.mk_nuprl_lambda_term id term)
				    term
				    vars)
		   end))
  | toNuprlTerm context (term as A.N {kind = (A.ID, A.ID_TYCON),  label, value, regions, parents, children = []}) = raise EH.Unimplemented ""
  | toNuprlTerm context (term as A.N {kind = (A.ID, A.ID_TYVAR),  label, value, regions, parents, children = []}) = raise EH.Unimplemented ""

  (* EXP *)
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_OR),       label, value, regions, parents, children = [e1, e2]})      =
    let val (nt1, pr1) = toNuprlTerm context e1
	val (nt2, pr2) = toNuprlTerm context e2
    in (NT.mk_nuprl_bor_term nt1 nt2,
	NT.mk_nuprl_bor_term pr1 pr2)
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_AND),      label, value, regions, parents, children = [e1, e2]})      =
    let val (nt1, pr1) = toNuprlTerm context e1
	val (nt2, pr2) = toNuprlTerm context e2
    in (NT.mk_nuprl_band_term nt1 nt2,
	NT.mk_nuprl_band_term pr1 pr2)
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_TYPED),    label, value, regions, parents, children = [e, t]})        =
    raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_LAMBDA),   label, value, regions, parents, children = [p, e]})        =
    let val id       = getIdPat 4 p
	val context' = add_bound_context context id
	val (nt, pr) = toNuprlTerm context' e
    in (NT.mk_nuprl_lambda_term id nt,
	NT.mk_nuprl_lambda_term id pr)
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_ITE),      label, value, regions, parents, children = [e1, e2, e3]})  =
    let val (nt1, pr1) = toNuprlTerm context e1
	val (nt2, pr2) = toNuprlTerm context e2
	val (nt3, pr3) = toNuprlTerm context e3
    in (NT.mk_nuprl_ite_term nt1 nt2 nt3,
	NT.mk_nuprl_ite_term pr1 pr2 pr3)
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_ATEXP),    label, value, regions, parents, children = [ae]})          =
    toNuprlTerm context ae
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_APP),      label, value, regions, parents, children = [f, x]})        =
    let val (t, ts) = flattenExpApp f
	val nts = map (toNuprlTerm context) (ts @ [x])
    in case t of
	   A.N {kind = (A.EXP, A.EXP_ATEXP), label, value, regions, parents, children = [atexp]} =>
	   (case atexp of
		A.N {kind = (A.ATEXP, A.ATEXP_ID), label, value, regions, parents, children = [v]} =>
		let val id  = A.getIdIdent v
		    val lab = A.getLabel v
		in case get_arity_context context id of
		       NONE =>
		       foldl (fn ((nt_arg, pr_arg), (nt_func, pr_func)) =>
				 (NT.mk_nuprl_apply_term nt_func nt_arg,
				  NT.mk_nuprl_apply_term pr_func pr_arg))
			     (toNuprlTerm context v)
			     nts
		     | SOME (idname, lst) =>
		       if is_param_context context id
		       then if List.null lst
			    then foldl (fn ((nt_arg, pr_arg), (nt_func, pr_func)) =>
					   (NT.mk_nuprl_apply_term nt_func nt_arg,
					    NT.mk_nuprl_apply_term pr_func pr_arg))
				       (double (NT.mk_nuprl_variable_term id))
				       nts
			    else raise EH.Impossible ""
		       else let val n  = List.length lst
				val ln = List.length nts
				val (params, args, bound) =
				    if ln >= n
				    then let val list1 = List.take (nts, n)
					     val list2 = List.drop (nts, n)
					 in (list1, list2, [])
					 end
				    else let val vars = List.tabulate (n - ln, fn _ => A.newId ())
					     (*val vars = map (fn (id, _) => A.newIdId id) (List.drop (lst, ln))*)
					     val nvars = map (fn v => double (NT.mk_nuprl_variable_term v)) vars
					 in (nts @ nvars, [], vars)
					 end
				val b_terms =
				    map (fn ((id, prms), (nt_arg, pr_arg)) =>
					    let val m = List.length prms
						fun toPair arg =
						    let val (bounds, body) = NT.dest_lambdas arg
							val bl = List.length bounds
						    in if bl >= m
						       then let val l1 = List.take (bounds, m)
								val l2 = List.drop (bounds, m)
							    in (l1, NT.mk_nuprl_lambdas_term l2 body)
							    end
						       else let val vars = List.tabulate (m - bl, fn _ => A.newId ())
								val term = foldl (fn (arg, term) =>
										     NT.mk_nuprl_apply_term term (NT.mk_nuprl_variable_term arg))
										 body
										 vars
							    in (bounds @ vars, term)
							    end
						    end
					    in (toPair nt_arg, toPair pr_arg)
					    end)
					(ListPair.zip (lst, params))
				val name = case EN.getNameAtLab (getUnifEnv ()) lab of
					       NONE => idname
					     | SOME name => name
				val (b_nt_terms, b_pr_terms) = ListPair.unzip b_terms
				val nt_term  = NT.mk_nuprl_term (name, []) b_nt_terms
				val pr_term  = NT.mk_nuprl_term (name, []) b_pr_terms
				val pr_term' = class2prog context lab pr_term
				val (nt_app, pr_app) =
				    foldl (fn ((nt_arg, pr_arg), (nt_func, pr_func)) =>
					      (NT.mk_nuprl_apply_term nt_func nt_arg,
					       NT.mk_nuprl_apply_term pr_func pr_arg))
					  (nt_term, pr_term')
					  args
				val (nt_lam, pr_lam) =
				    foldr (fn (f, (nt_term, pr_term)) =>
					      (NT.mk_nuprl_lambda_term f nt_term,
					       NT.mk_nuprl_lambda_term f pr_term))
					  (nt_app, pr_app)
					  bound
			    in (nt_lam, pr_lam)
			    end handle Subscript => raise EH.Impossible ""
		end
	      | _ => foldl (fn ((nt_arg, pr_arg), (nt_func, pr_func)) =>
			       (NT.mk_nuprl_apply_term nt_func nt_arg,
				NT.mk_nuprl_apply_term pr_func pr_arg))
			   (toNuprlTerm context atexp)
			   nts)
	 | _ => foldl (fn ((nt_arg, pr_arg), (nt_func, pr_func)) =>
			  (NT.mk_nuprl_apply_term nt_func nt_arg,
			   NT.mk_nuprl_apply_term pr_func pr_arg))
		      (toNuprlTerm context t)
		      nts
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_OP),       label, value, regions, parents, children = [e1, e2]})      =
    let val (nt1, pr1) = toNuprlTerm context e1
	val (nt2, pr2) = toNuprlTerm context e2
	fun addTypCls f =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		SOME ity =>
		let val t1 = toNuprlIty_em 8 false context (E.stripIty ity)
		    val (_, t2) = NT.dest_function t1
		    val (_, t3) = NT.dest_function t2
		    val (info, es, e, typ) = NT.dest_eclass t3
		(* NOTE: we should check that es and e do not occur in typ. *)
		in f typ
		end
	      | NONE => raise EH.Impossible ("no type for the " ^ value ^ " operator")
    in case value of
	   "+"     => (NT.mk_nuprl_add_term       nt1 nt2, NT.mk_nuprl_add_term       pr1 pr2)
	 | "-"     => (NT.mk_nuprl_subtract_term  nt1 nt2, NT.mk_nuprl_subtract_term  pr1 pr2)
	 | "*"     => (NT.mk_nuprl_multiply_term  nt1 nt2, NT.mk_nuprl_multiply_term  pr1 pr2)
	 | "/"     => (NT.mk_nuprl_divide_term    nt1 nt2, NT.mk_nuprl_divide_term    pr1 pr2)
	 | "="     => (case EN.getItyAtLab (getUnifEnv ()) label of
			   (* NOTE: this use of labels assumes that we never change
			    * the label of a EXP_OP. *)
			   NONE => raise EH.Impossible "We shoud have generated a type for this equality"
			 | SOME ity =>
			   let val ity1   = E.stripIty ity
			       val ity2   = Option.valOf (E.destDeqType ity1)
				   handle _ => raise EH.Impossible ""
			       val deq    = toDeq context ity2
			       val nt_app = NT.mk_nuprl_apply_term deq nt1
			       val pr_app = NT.mk_nuprl_apply_term deq pr1
			   in (NT.mk_nuprl_apply_term nt_app nt2,
			       NT.mk_nuprl_apply_term pr_app pr2)
			   end)
	 | "<>"    => (case EN.getItyAtLab (getUnifEnv ()) label of
			   NONE => raise EH.Impossible "We shoud have generated a type for this equality"
			 | SOME ity =>
			   let val ity1   = E.stripIty ity
			       val ity2   = Option.valOf (E.destDeqType ity1)
				   handle _ => raise EH.Impossible ""
			       val deq     = toDeq context ity2
			       val nt_app1 = NT.mk_nuprl_apply_term deq     nt1
			       val nt_app2 = NT.mk_nuprl_apply_term nt_app1 nt2
			       val pr_app1 = NT.mk_nuprl_apply_term deq     pr1
			       val pr_app2 = NT.mk_nuprl_apply_term pr_app1 pr2
			   in (NT.mk_nuprl_bnot_term nt_app2,
			       NT.mk_nuprl_bnot_term pr_app2)
			   end)
	 | "<="    => let val nt_leq  = NT.mk_nuprl_lt_int_term nt1 nt2
			  val pr_leq  = NT.mk_nuprl_lt_int_term pr1 pr2
			  val deq     = NT.mk_nuprl_int_deq_term
			  val nt_app1 = NT.mk_nuprl_apply_term deq     nt1
			  val nt_app2 = NT.mk_nuprl_apply_term nt_app1 nt2
			  val pr_app1 = NT.mk_nuprl_apply_term deq     pr1
			  val pr_app2 = NT.mk_nuprl_apply_term pr_app1 pr2
		      in (NT.mk_nuprl_bor_term nt_leq nt_app2,
			  NT.mk_nuprl_bor_term pr_leq pr_app2)
		      end
	 | ">="    => let val nt_leq  = NT.mk_nuprl_lt_int_term nt2 nt1
			  val pr_leq  = NT.mk_nuprl_lt_int_term pr2 pr1
			  val deq     = NT.mk_nuprl_int_deq_term
			  val nt_app1 = NT.mk_nuprl_apply_term deq     nt1
			  val nt_app2 = NT.mk_nuprl_apply_term nt_app1 nt2
			  val pr_app1 = NT.mk_nuprl_apply_term deq     pr1
			  val pr_app2 = NT.mk_nuprl_apply_term pr_app1 pr2
		      in (NT.mk_nuprl_bor_term nt_leq nt_app2,
			  NT.mk_nuprl_bor_term pr_leq pr_app2)
		      end
	 | "<"     => (NT.mk_nuprl_lt_int_term           nt1 nt2, NT.mk_nuprl_lt_int_term           pr1 pr2)
	 | ">"     => (NT.mk_nuprl_lt_int_term           nt2 nt1, NT.mk_nuprl_lt_int_term           pr2 pr1)
	 | "."     => (NT.mk_nuprl_cons_term             nt1 nt2, NT.mk_nuprl_cons_term             pr1 pr2)
	 | "++"    => (NT.mk_nuprl_concat_term           nt1 nt2, NT.mk_nuprl_concat_term           pr1 pr2)
	 | "@"     => (NT.mk_nuprl_class_at_term         nt1 nt2, addTypCls (fn typ => NT.mk_nuprl_at_prc_term       typ pr1 pr2))
	 | "||"    => (NT.mk_nuprl_parallel_class_term   nt1 nt2, addTypCls (fn typ => NT.mk_nuprl_parallel_prc_term typ pr1 pr2))
	 | "until" => (NT.mk_nuprl_until_class_term      nt1 nt2, addTypCls (fn typ => NT.mk_nuprl_until_prc_term    typ pr1 pr2))
	 | "?"     =>
	   (if NT.opid_of_term nt1 = "primed-class"
	    then NT.mk_nuprl_primed_class_opt_term (NT.dest_primed_class nt1) nt2
	    else NT.mk_nuprl_class_opt_term nt1 nt2,
	    if NT.opid_of_term pr1 = "prior-prc"
	    then let val (t,X) = NT.dest_prior_prc pr1
		     fun F typ = NT.mk_nuprl_prior_init_prc_term typ X pr2
		 in addTypCls F
		 end
	    else NT.mk_nuprl_class_opt_term pr1 pr2 (* opt-prc ? *))
	 | "??"    => (NT.mk_nuprl_class_opt_class_term nt1 nt2,
		       NT.mk_nuprl_class_opt_class_term pr1 pr2)
	 | _       =>
	   (case get_arity_context context value of
		NONE => (* the id is not a constant, it is a user function *)
		(case getUserFP (is_bound_context context) value of
		     (* We sometimes have to add parameters to the argument list
		      * of a used function, if it does use the parameters. *)
		     NONE => double (NT.mk_nuprl_variable_term value)
		   | SOME params =>
		     let val a_terms = map NT.mk_nuprl_variable_term params
			 val _       = addDecFPList params
			 val func    = mk_nuprl_user_term value a_terms
		     in (NT.mk_nuprl_apply_term (NT.mk_nuprl_apply_term func nt1) nt2,
			 NT.mk_nuprl_apply_term (NT.mk_nuprl_apply_term func pr1) pr2)
		     end)
	      | _ => raise EH.Impossible "a constant or a parameter cannot be infix")
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_LET),      label, value, regions, parents, children = [b, e]})        =
    (case b of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e']} =>
	 let val id         = A.getIdIdent f
	     val (nte, pre) = toNuprlTerm context e'
	     val context'   = add_bound_context context id
	     val (nt, pr)   = toNuprlTerm context' e
	 in (NT.mk_nuprl_let_term id nte nt,
	     NT.mk_nuprl_let_term id pre pr)
	 end
       | A.N {kind = (A.BIND, A.BIND_PAT), label, value = v, regions, parents, children = [p, e']} =>
	 (case p of
	      A.N {kind = (A.PAT, A.PAT_ATPAT), label, value, regions, parents, children = [atpat]} =>
	      (case atpat of
		   A.N {kind = (A.ATPAT, A.ATPAT_TUPLE), label, value, regions, parents, children} =>
		   if List.length children <= 7
		   then let val vars       = map (getIdPat 5) children
			    val (nte, pre) = toNuprlTerm context e'
			    val context'   = add_bounds_context context vars
			    val (nt, pr)   = toNuprlTerm context' e
			in (NT.mk_nuprl_spreadn_term nte (vars, nt),
			    NT.mk_nuprl_spreadn_term pre (vars, pr))
			end
		   else raise EH.Impossible minStr
		 | A.N {kind = (A.ATPAT, A.ATPAT_ID), label, value, regions, parents, children = [ident]} =>
		   let val id         = A.getIdIdent ident
		       val (nte, pre) = toNuprlTerm context e'
		       val context'   = add_bound_context context id
		       val (nt, pr)   = toNuprlTerm context' e
		   in (NT.mk_nuprl_let_term id nte nt,
		       NT.mk_nuprl_let_term id pre pr)
		   end
		 | atpat => (print (A.toString atpat ^ "\n"); raise EH.Impossible A.wrongFormat))
	    | pat => (print (A.toString pat ^ "\n"); raise EH.Impossible minStr))
       | _ => raise EH.Impossible minStr)
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_CLASS),    label, value, regions, parents, children = [b, e]})        = raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_LETR),     label, value, regions, parents, children = [b, e]})        =
    (case b of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e']} =>
	 let val fid        = A.getIdIdent f
	     val context'   = add_bound_context context fid
	     val (nte, pre) = toNuprlTerm context' e'
	     val (nt, pr)   = toNuprlTerm context' e
	     fun toRec e1 e2 =
		 if NT.is_nuprl_lambda_term e1
		 then let val (x, B) = NT.dest_lambda e1
			  (* gind: letrec f x = B in f *)
			  val gind = NT.mk_nuprl_genrec_term (x, fid, B)
		      in if NT.is_nuprl_variable_term e2
			    andalso
			    NT.dest_variable e2 = fid
			 then gind
			 else let val lam = NT.mk_nuprl_lambda_term fid e2
			      (* lam: \f. e *)
			      (* term: (\f. e) (letrec f x = B in f) *)
			      in NT.mk_nuprl_apply_term lam gind
			      end
		      end
		 else (print (A.toString e'); raise EH.Impossible "recursive definition is not a function")
	 in (toRec nte nt, toRec pre pr)
	 end
       | _ => raise EH.Impossible minStr)
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_WHERE),    label, value, regions, parents, children = [e, b]})        =
    raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_BINDING),  label, value, regions, parents, children = [e1, p, e2]})   =
    (case A.getIdPat p of
	 SOME ident =>
	 let val (nt1, pr1) = toNuprlTerm context e1
	     val id         = A.getIdIdent ident
	     val context'   = add_bound_context context id
	     val (nt2, pr2) = toNuprlTerm context' e2
  	     val (A, B)     =
		 case EN.getItyAtLab (getUnifEnv ()) label of
		     NONE     => raise EH.Impossible "bind combinator without type"
		   | SOME ity =>
		     let val typ = toNuprlIty_em 9 false context (E.stripIty ity)
			 val (left, right) = NT.dest_function typ
			 val (info, es, e, Xtyp) = NT.dest_eclass right
		     in (left, Xtyp)
		     end
	 in (NT.mk_nuprl_bind_class_term nt1 (id, nt2),
	     NT.mk_nuprl_bind_prc_term A B pr1 (NT.mk_nuprl_lambda_term id pr2))
	 end
       | NONE => (print (A.toString p ^ "\n"); raise EH.Impossible minStr))
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_MBIND),    label, value, regions, parents, children = [e1, e2]})      = raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_COMP),     label, value, regions, parents, children = (exp :: exps)}) =
    let val (nt, pr) = toNuprlTerm context exp
	val lst = map (toNuprlTerm context) exps
	val typ =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE => raise EH.Impossible "we should have generated a type for simple/rec combinators"
	      | SOME ity => toNuprlIty_em 14 true context (E.stripIty ity)
	val bp  = String.isSubstring "P" value
	val bf  = String.isSubstring "F" value
	val bc  = String.isSubstring "C" value
	val bo  = String.isSubstring "O" value
	fun getOpt nts =
	    if bo andalso not (List.null nts)
	    then let val rev = List.rev nts
		     val (e_last, p_last) = List.hd rev
		     val firsts = List.rev (List.tl rev)
		 in if A.isExpBag (List.last exps)
		    then let val id    = A.newIdIdSet (get_all_ids_list [e_last, p_last]) "l"
			     val e_lam = NT.mk_nuprl_lambda_term id e_last
			     val p_lam = NT.mk_nuprl_lambda_term id p_last
			 in (firsts, SOME (e_lam, p_lam))
			 end
		    else (firsts, SOME (e_last, p_last))
		 end
	    else (nts, NONE)
	val (lst', opt) = getOpt lst
    in preToNuprlCombComp (nt, pr) lst' bp opt typ bf bc
    end
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_QUOT),     label, value, regions, parents, children = [e]})           = toNuprlTerm context e
  | toNuprlTerm context (term as A.N {kind = (A.EXP, A.EXP_CASE),     label, value, regions, parents, children = [e, m]})        =
    let val (e_term, p_term) = toNuprlTerm context e
    in case A.caseMatch m of
	   A.CASE_LIST (exp1, (id1, id2, exp2)) =>
	   let val (nt1, pr1) = toNuprlTerm context exp1
	       val context'   = add_bounds_context context [id1, id2]
	       val (nt2, pr2) = toNuprlTerm context' exp2
	       fun toT t      = (id1, id2, A.newId (), t)
	   in (NT.mk_nuprl_list_ind_term e_term nt1 (toT nt2),
	       NT.mk_nuprl_list_ind_term p_term pr1 (toT pr2))
	   end
	 | A.CASE_BAG (NONE, (pat, exp1), exp2) =>
	   let val (nt1, pr1) =
		   case A.getIdPat pat of
		       SOME id =>
		       let val nid      = A.getIdIdent id
			   val context' = add_bound_context context nid
			   val (nt, pr) = toNuprlTerm context' exp1
			   val nt_lam   = NT.mk_nuprl_lambda_term nid nt
			   val pr_lam   = NT.mk_nuprl_lambda_term nid pr
		       in (fn arg => NT.mk_nuprl_apply_term nt_lam arg,
			   fn arg => NT.mk_nuprl_apply_term pr_lam arg)
		       end
		     | NONE => raise EH.Impossible "We have transformed the pat of a msg casepat into an id."
	       val (nt2, pr2) = toNuprlTerm context exp2
	       fun toTerm t1 t2 t =
		   let val b     = NT.is_nuprl_variable_term t
		       val mid   = if b then NT.dest_variable t else A.newId ()
		       val tmid  = NT.mk_nuprl_variable_term mid
		       val arg   = NT.mk_nuprl_bag_only_term tmid
		       val app   = t1 arg
		       val bsize = NT.mk_nuprl_bag_size_term tmid
		       val none  = NT.mk_nuprl_integer_term 1
		       val bone  = NT.mk_nuprl_eq_int_term bsize none
		       val ite   = NT.mk_nuprl_ite_term bone app t2
		   in if b
		      then ite
		      else let val lam = NT.mk_nuprl_lambda_term mid ite
			   in NT.mk_nuprl_apply_term lam t
			   end
		   end
	   in (toTerm nt1 nt2 e_term, toTerm pr1 pr2 p_term)
	   end
	 | A.CASE_BAG (SOME exp1, (pat, exp2), exp3) =>
	   let val (nt1, pr1) = toNuprlTerm context exp1
	       val (nt2, pr2) =
		   case A.getIdPat pat of
		       SOME id =>
		       let val nid      = A.getIdIdent id
			   val context' = add_bound_context context nid
			   val (nt, pr) = toNuprlTerm context' exp2
			   val nt_lam   = NT.mk_nuprl_lambda_term nid nt
			   val pr_lam   = NT.mk_nuprl_lambda_term nid pr
		       in (fn arg => NT.mk_nuprl_apply_term nt_lam arg,
			   fn arg => NT.mk_nuprl_apply_term pr_lam arg)
		       end
		     | NONE => raise EH.Impossible "We have transformed the pat of a msg casepat into an id."
	       val (nt3, pr3) = toNuprlTerm context exp3
	       fun toTerm t1 t2 t3 t =
		   let val b     = NT.is_nuprl_variable_term t
		       val mid   = if b then NT.dest_variable t else A.newId ()
		       val tmid  = NT.mk_nuprl_variable_term mid
		       val bnull = NT.mk_nuprl_bag_null_term tmid
		       val arg   = NT.mk_nuprl_bag_only_term tmid
		       val app   = t2 arg
		       val bsize = NT.mk_nuprl_bag_size_term tmid
		       val none  = NT.mk_nuprl_integer_term 1
		       val bone  = NT.mk_nuprl_eq_int_term bsize none
		       val ite1  = NT.mk_nuprl_ite_term bone app t3
		       val ite2  = NT.mk_nuprl_ite_term bnull t1 ite1
		   in if b
		      then ite2
		      else let val lam = NT.mk_nuprl_lambda_term mid ite2
			   in NT.mk_nuprl_apply_term lam t
			   end
		   end
	   in (toTerm nt1 nt2 nt3 e_term, toTerm pr1 pr2 pr3 p_term)
	   end
	 | A.CASE_MSG (lst, exp) =>
	   (* The exp is in the wildcard case.
	    * Each element in lst is of the form (atom, pat, exp, labM) for each
	    * `MSG(atom,pat) =>^labM exp' branch *)
	   let val b     = NT.is_nuprl_variable_term e_term
	       (* if e_term is a variable, then p_term is also a variable,
		* and it is the same variable.  (It would be better to check
		* it though.) *)
	       val mid   = if b then NT.dest_variable e_term else A.newId ()
	       val tmid  = NT.mk_nuprl_variable_term   mid
	       val t_hdr = NT.mk_nuprl_msg_header_term tmid
	       val t_typ = NT.mk_nuprl_msg_type_term   tmid
	       val t_bdy = NT.mk_nuprl_msg_body_term   tmid
	       val (t_exp, p_exp) = toNuprlTerm context exp
	       val nid   = A.newIdIdSet (get_all_ids_list [t_exp, p_exp]) "_"
	       val ltexp = NT.mk_nuprl_lambda_term nid t_exp
	       val lpexp = NT.mk_nuprl_lambda_term nid p_exp
	       val (term, prog) =
		   foldr (fn ((atom, pat, exp, labM), (term, prog)) =>
			     case (A.getIdPat pat, atom) of
				 (SOME id, A.N {kind = (A.ATOMS, A.ATOMS_ATOMS), label, value, regions, parents, children = []}) =>
				 let val n_atoms = toNuprlListAtoms (A.splitAtoms value)
				     val n_eq1   = NT.mk_nuprl_name_eq_term t_hdr n_atoms
				     val n_ity   = case EN.getItyAtLab (getUnifEnv ()) labM of
						       NONE => raise EH.Impossible "we should have generated a type for each pattern message"
						     | SOME ity => toNuprlIty_em 3 false context (E.stripIty ity)
				     val n_eq2   = NT.mk_nuprl_msg_has_type_term tmid n_ity
				     val n_eq    = NT.mk_nuprl_band_term n_eq1 n_eq2
				     val (t, p)  = toNuprlTerm context exp
				     val tlam    = NT.mk_nuprl_lambda_term (A.getIdIdent id) t
				     val plam    = NT.mk_nuprl_lambda_term (A.getIdIdent id) p
				 in (NT.mk_nuprl_ite_term n_eq tlam term,
				     NT.mk_nuprl_ite_term n_eq plam term)
				 end
			       | (NONE, _) => raise EH.Impossible "We have transformed the pat of a msg casepat into an id."
			       | _ => raise EH.Impossible "We have transformed the atoms into the slanted quoted forms.")
			 (ltexp, lpexp)
			 lst
	       val tapp = NT.mk_nuprl_apply_term term t_bdy
	       val papp = NT.mk_nuprl_apply_term prog t_bdy
	   in if b
	      then (tapp, papp)
	      else let val tlam = NT.mk_nuprl_lambda_term mid tapp
		       val plam = NT.mk_nuprl_lambda_term mid papp
		   in (NT.mk_nuprl_apply_term tlam e_term,
		       NT.mk_nuprl_apply_term plam p_term)
		   end
	   end
	 | A.CASE_OTH => (print (A.toString m ^ "\n" ^ A.export m); raise EH.Unimplemented "")
    end

  (* ATEXP *)
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_ID),      label, value, regions, parents, children = [i]})      = toNuprlTerm context i
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_SCON),    label, value, regions, parents, children = [sc]})     = toNuprlTerm context sc
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_TUPLE),   label, value, regions, parents, children = []})       = double (NT.mk_nuprl_it_term)
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_TUPLE),   label, value, regions, parents, children = [e1, e2]}) =
    let val (nt1, pr1) = toNuprlTerm context e1
	val (nt2, pr2) = toNuprlTerm context e2
    in (NT.mk_nuprl_pair_term nt1 nt2,
	NT.mk_nuprl_pair_term pr1 pr2)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_TUPLE),   label, value, regions, parents, children})            = raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_PAREN),   label, value, regions, parents, children = [x]})      = toNuprlTerm context x
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_LIST),    label, value, regions, parents, children})            = raise EH.Impossible minStr
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_BAG),     label, value, regions, parents, children})            =
    (case children of
	 [x] =>
	 let val (nt, pr) = toNuprlTerm context x
	 in (NT.mk_nuprl_single_bag_term nt, NT.mk_nuprl_single_bag_term pr)
	 end
       | _   => foldr (fn (x, (nt_term, pr_term)) =>
			  let val (nt, pr) = toNuprlTerm context x
			  in (NT.mk_nuprl_cons_bag_term nt nt_term,
			      NT.mk_nuprl_cons_bag_term pr pr_term)
			  end)
		      (double NT.mk_nuprl_empty_bag_term)
		      children)
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_PRIOR),   label, value, regions, parents, children = [e]})      =
    let val (nt, pr) = toNuprlTerm context e
	val typ =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		SOME ity => toNuprlIty_em 13 false context (E.stripIty ity)
	      | NONE     => raise EH.Impossible "We should have generated a type for prior"
    in (NT.mk_nuprl_primed_class_term nt,
	NT.mk_nuprl_prior_prc_term typ pr)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_BASE),    label, value, regions, parents, children = [e1, e2]}) =
    let val (nt1, pr1) = toNuprlTerm context e1
	val (nt2, pr2) = toNuprlTerm context e2
    in (NT.mk_nuprl_general_base_class_term nt1 nt2,
	NT.mk_nuprl_general_base_class_term pr1 pr2)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_ABASE),   label, value, regions, parents, children = [a, e]}) =
    let val hdrs =
	    case a of
		A.N {kind = (A.ATOMSL, A.ATOMSL_S), label, value, regions, parents, children} =>
		NT.mk_nuprl_finite_list_term (map toAtomList children)
	      | _ => raise EH.Impossible ""
	val (nt_locs, pr_locs) = toNuprlTerm context e
    in (NT.mk_nuprl_base_locs_headers_term hdrs nt_locs,
	NT.mk_nuprl_base_locs_headers_term hdrs pr_locs)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_ANY),     label, value, regions, parents, children = [e]})      = raise EH.Impossible "cannot convert an any term"
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_MSG),     label, value, regions, parents, children = [a, e]})   =
    (* TODO: We need type information here, if we know that the piece of code
     * is typable and if we got the generated environment then, we can just
     * lookup the type associated to the atoms of the message.
     * We can't actually do that because we can have a type message of the form
     * (``foo`` : 'a List) and the type of the message we're looking at is of
     * type Int List. *)
    let val nta        = toAtomList a
	val (nte, pre) = toNuprlTerm context e
    in case EN.getItyAtLab (getUnifEnv ()) label of
	   NONE     => raise EH.Impossible "We should have generated a type for this message"
	 | SOME ity =>
	   let val n_ity = toNuprlIty_em 4 false context (E.stripIty ity)
	   in (NT.mk_nuprl_make_msg_term nta n_ity nte,
	       NT.mk_nuprl_make_msg_term nta n_ity pre)
	   end
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_ONCE), label, value, regions, parents, children = [e]}) =
    let val (nte, pre) = toNuprlTerm context e
	val typ =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for once classes"
	      | SOME ity => toNuprlIty_em 10 false context (E.stripIty ity)
    in (NT.mk_nuprl_once_class_term nte,
	NT.mk_nuprl_once_prc_term typ pre)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_SENDOC), label, value, regions, parents, children = [e]}) =
    let val (nte, pre) = toNuprlTerm context e
	val typ =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for once classes"
	      | SOME ity => toNuprlIty_em 11 false context (E.stripIty ity)
    in (NT.mk_nuprl_send_once_loc_class_term nte,
	NT.mk_nuprl_send_once_loc_prc_term typ pre)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_ONLOC), label, value, regions, parents, children = [e]}) =
    let val (nte, pre) = toNuprlTerm context e
	val typ =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for once classes"
	      | SOME ity => toNuprlIty_em 12 false context (E.stripIty ity)
    in (NT.mk_nuprl_on_loc_class_term nte,
	NT.mk_nuprl_on_loc_prc_term typ pre)
    end
  | toNuprlTerm context (term as A.N {kind = (A.ATEXP, A.ATEXP_MINUS), label, value, regions, parents, children = [e]}) =
    let val (nte, pre) = toNuprlTerm context e
    in (NT.mk_nuprl_minus_term nte,
	NT.mk_nuprl_minus_term pre)
    end

  (* PROP *)
  | toNuprlTerm context (term as A.N {kind = (A.PROP, A.PROP_EXP), label, value, regions, parents, children = [exp]}) =
    let val (nte, pre) = toNuprlTerm context exp
    in (nte, pre)
    end

  (* PARAM *)
  | toNuprlTerm context (term as A.N {kind = (A.PARAM, _),      label, value, regions, parents, children}) = raise EH.Unimplemented ""
  (* TYPEVARSEQ *)
  | toNuprlTerm context (term as A.N {kind = (A.TYPEVARSEQ, _), label, value, regions, parents, children}) = raise EH.Unimplemented ""
  (* TYPESEQ *)
  | toNuprlTerm context (term as A.N {kind = (A.TYPESEQ, _),    label, value, regions, parents, children}) = raise EH.Unimplemented ""
  (* DOTS *)
  | toNuprlTerm context (term as A.N {kind = (A.DOTS, _),       label, value, regions, parents, children}) = raise EH.Unimplemented ""

  (* wrong format *)
  | toNuprlTerm context term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun toNuprlAtomElt context atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_BASE), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val term = NT.mk_nuprl_base_headers_msg_val_term atm typ
	val prog = NT.mk_nuprl_base_prc_term atm typ
    in ((value, id, term, dummy_dot_term), prog)
    end
  | toNuprlAtomElt context atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_BASEL), label, value, regions, parents, children = [ident, exp]}) =
    let val id       = A.getIdIdent ident
	val (ne, pe) = toNuprlTerm context exp
	val term     = NT.mk_nuprl_base_headers_msg_val_loc_term atm ne typ
	val prog     = NT.mk_nuprl_base_at_prc_term atm typ pe
    in ((value, id, term, exp), prog)
    end
  | toNuprlAtomElt context atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_MSG), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val set  = get_all_ids_list [atm, typ]
	val v    = A.newIdIdSet set "z"
	val tv   = NT.mk_nuprl_variable_term v
	(* msg: <atom, type, value> *)
	val msg  = NT.mk_nuprl_make_msg_term atm typ tv
	(* msg: \value. <atom, type, value> *)
	val term = NT.mk_nuprl_lambda_term v msg
    in ((value, id, term, dummy_dot_term), term)
    end
  | toNuprlAtomElt context atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_SEND), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val set  = get_all_ids_list [atm, typ]
	val v    = A.newIdIdSet set "z"
	val l    = A.newIdIdSet set "l"
	val tv   = NT.mk_nuprl_variable_term v
	val tl   = NT.mk_nuprl_variable_term l
	(* msg: <atom, type, value> *)
	val msg  = NT.mk_nuprl_make_msg_term atm typ tv
	(* pair: <loc, <atom, type, value>> *)
	val pair = NT.mk_nuprl_pair_term tl msg
	(* lam: \value. <loc, <atom, type, value>> *)
	val lam  = NT.mk_nuprl_lambda_term v pair
	(* term: \loc. \value. <loc, <atom, type, value>> *)
	val term = NT.mk_nuprl_lambda_term l lam
    in ((value, id, term, dummy_dot_term), term)
    end
  | toNuprlAtomElt context atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_BCAST), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val set  = get_all_ids_list [atm, typ]
	val v    = A.newIdIdSet set "z"
	val l    = A.newIdIdSet set "l"
	val ls   = A.newIdIdSet set "locs"
	val tv   = NT.mk_nuprl_variable_term v
	val tl   = NT.mk_nuprl_variable_term l
	val tls  = NT.mk_nuprl_variable_term ls
	(* msg: <atom, type, value> *)
	val msg  = NT.mk_nuprl_make_msg_term atm typ tv
	(* pair: <loc, <atom, type, value>> *)
	val pair = NT.mk_nuprl_pair_term tl msg
	(* llam:  \loc. <loc, <atom, type, value>> *)
	val llam = NT.mk_nuprl_lambda_term l pair
	(* bmap: bag-map(\loc. <loc, <atom, type, value>>,locs) *)
	val bmap = NT.mk_nuprl_bag_map_term llam tls
	(* lam: \value. bag-map(\loc. <loc, <atom, type, value>>,locs) *)
	val lam  = NT.mk_nuprl_lambda_term v bmap
	(* term: \locs. \value. bag-map(\loc. <loc, <atom, type, value>>,locs) *)
	val term = NT.mk_nuprl_lambda_term ls lam
    in ((value, id, term, dummy_dot_term), term)
    end
  | toNuprlAtomElt context atm typ term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* send = \l. \v. (l, MSG(``hdr``, v)) *)
(* broadcast = \locs. \v. map (\l. (l, MSG(``hdr``, v))) locs *)

fun toNuprlAtomElts context atm typ (term as A.N {kind = (A.ATOMELTS, A.ATOMELTS_ELTS), label, value, regions, parents, children}) =
    foldr (fn (elt, (decs, progs)) =>
	      let val ((v, i, t, e), prog) = toNuprlAtomElt context atm typ elt
		  val (ab, th, rest) = genAbstractionAndWF' 1 context v i t e []
		  val _     = resetDecFP ()
	      in (ab :: th :: rest @ decs, (get_user_name i, prog) :: progs)
	      end)
	  ([], [])
	  children
  | toNuprlAtomElts context atm typ term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* TATOM *)
fun toNuprlTAtom context (term as A.N {kind = (A.TATOM, A.TATOM_G), label, value, regions, parents, children = [atoms, typ, elts]}) =
    let val atm  = toAtomList atoms
	val typ  =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for this message"
	      | SOME ity => toNuprlIty_em 7 false context (E.stripIty ity)
	val _    = addHeader (value, atm)
    in toNuprlAtomElts context atm typ elts
    end
  | toNuprlTAtom context term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun toNuprlDoc [] = NT.mk_nuprl_itext_nil_term
  | toNuprlDoc (doc :: docs) =
    let val text = NT.mk_nuprl_itext_term (protect doc)
    in NT.mk_nuprl_itext_cons_term
	   text
	   (NT.mk_nuprl_itext_cons_term
		NT.mk_nuprl_inewline_term
		(toNuprlDoc docs))
    end

fun addDoc [] term = term
  | addDoc [""] term = term
  | addDoc doc term =
    let val prop = NT.mk_nuprl_iproperty_term "doc" (toNuprlDoc doc)
	val cons = NT.mk_nuprl_icons_cons_term prop NT.mk_nuprl_icons_nil_term
    in NT.mk_nuprl_iinclude_properties_term cons term
    end

fun toNuprlData context data n len (A.N {kind = (A.DATA, A.DATA_CONS), label, value, regions, parents, children = [id, typ]}) =
    let val typ  =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for this message"
	      | SOME ity => toNuprlIty_em 17 false context (E.stripIty ity)
	val name   = A.getIdIdent id
	val id_is  = "is_"  ^ name
	val id_get = "get_" ^ name
	val tdata  =
	    case is_data_context context data of
		SOME prms =>
		let val a_terms = map NT.mk_nuprl_variable_term prms
		    val _ = addDecFPList prms
		in mk_nuprl_user_term name a_terms
		end
	      | NONE => raise Fail "data"
	(*val typ1   = NT.mk_nuprl_fun_term typ tdata
	val typ2   = NT.mk_nuprl_fun_term tdata NT.mk_nuprl_bool_term
	val typ3   = NT.mk_nuprl_fun_term tdata typ*)
	val xt = NT.mk_nuprl_variable_term "x"
	fun set x m =
	    if len < 1 orelse m < 0
	    then raise Fail "toNuprlData:neg"
	    else if len = 1
	    then x
	    else if m = 0 andalso not (len - 1 = n) (* i.e., not last cons *)
	    then NT.mk_nuprl_inl_term x
	    else NT.mk_nuprl_inr_term (if m = 0 then x else set x (m - 1))
	fun get x m =
	    if len < 1 orelse m < 0
	    then raise Fail "toNuprlData:neg"
	    else if len = 1
	    then x
	    else if m = 0 andalso not (len - 1 = n)
	    then NT.mk_nuprl_outl_term x
	    else if m = 0
	    then NT.mk_nuprl_outr_term x
	    else get (NT.mk_nuprl_outr_term x) (m - 1)
	fun is x m =
	    if len < 1 orelse m < 0
	    then raise Fail "toNuprlData:neg"
	    else if len = 1
	    then NT.mk_nuprl_btrue_term
	    else if m = 0 andalso not (len - 1 = n)
	    then NT.mk_nuprl_isl_term x
	    else if m = 0
	    then NT.mk_nuprl_isr_term x
	    else let val b1 = NT.mk_nuprl_isr_term x
		     val b2 = is (NT.mk_nuprl_outr_term x) (m - 1)
		 in NT.mk_nuprl_band_term b1 b2
		 end
	val term1 = NT.mk_nuprl_lambda_term "x" (set xt n)
	val term2 = NT.mk_nuprl_lambda_term "x" (is  xt n)
	val term3 = NT.mk_nuprl_lambda_term "x" (get xt n)
	val (ab1, th1, rest1) = genAbstractionAndWF' 5 context "" name   term1 dummy_dot_term []
	val (ab2, th2, rest2) = genAbstractionAndWF' 6 context "" id_is  term2 dummy_dot_term []
	val app = fn x => NT.mk_nuprl_assert_term (NT.mk_nuprl_apply_term (#1 (toNuprlTerm context (A.idToExp id_is))) x)
	val (ab3, th3, rest3) = genAbstractionAndWF' 7 context "" id_get term3 dummy_dot_term [app]
	val _ = resetDecFP ()
	val decs  = [ab1, th1] @ rest1 @ [ab2, th2] @ rest2 @ [ab3, th3] @ rest3
	val progs = [(get_user_name name, term1), (get_user_name id_is, term2),(get_user_name id_get, term3)]
    in (decs, progs)
    end
  | toNuprlData context data n len term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* DECS *)
fun toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_LET), label, value, regions, parents, children = [b]}) =
    (case b of
	 A.N {kind = (A.BIND, A.BIND_DEC),  label, value, regions, parents, children = [f, a, e]} =>
	 if A.termIsEmParam a
	 then let val id       = A.getIdIdent f
		  val (ne, pe) = toNuprlTerm context e
		  val (ab, th, rest) = genAbstractionAndWF' 4 context value id ne e []
		  val ab'      = addDoc doc ab
		  val _        = resetDecFP ()
	      in (ab' :: th :: rest, [(get_user_name id, pe)], context, NONE)
	      end
	 else raise EH.Impossible minStr
       | _ => raise EH.Impossible minStr)
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_CLASS),   label, value, regions, parents, children = [b]}) = raise EH.Impossible minStr
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_LETR),    label, value, regions, parents, children = [b]}) =
    raise EH.Impossible minStr
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_CONS),    label, value, regions, parents, children = [i, a, t]}) =
    let val id = A.getIdIdent i
	val ar = getArityArgs a
	val _  = addOneId id
	val context1 = add_new_id_context context id
	val context2 = add_arity_context context1 id id ar
    in ([], [], context2, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_OCONS),   label, value, regions, parents, children = [i, a, t, v, s]}) =
    let val id = A.getIdIdent i
	val ar = getArityArgs a
	val _  = addOneId id
	val context1 = add_new_id_context context id
	val context2 = add_arity_context context1 id id ar
    in ([], [], context2, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_PARAM), label, value, regions, parents, children = [i, _]}) =
    let val id  = A.getIdIdent i
	val nb  = getNbOfId id
	val sch =
	    case EN.getScheme id (nb + 1) (getUnifEnv ()) of
		NONE => raise EH.Impossible ("Couldn't find " ^ id ^ "'s environment")
	      | SOME scheme => scheme
	val _  = addOneId id
	val context1 = add_new_id_context context id
	val context2 = add_arity_context context1 id id []
	val context3 = add_param_context context2 id sch
    in ([], [], context3, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_TYPARAM), label, value, regions, parents, children = [s, tc, _]}) =
    let val t1  = (*E.mk_type_typ1*) E.mk_type_type D.dummy_label
	val typ = foldr (fn (_, term) => E.mk_type_arrow (t1, term) D.dummy_label)
			t1
			(List.tabulate (getArityTyvarSeq s, fn _ => ()))
	val id  = A.getIdIdent tc
	val sch = E.mk_new_scheme [] typ
    in ([], [], add_param_context context id sch, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_ETPARAM), label, value, regions, parents, children = [s, tc, i, _]})   =
    let val tc   = A.getIdIdent tc
	val ar   = getArityTyvarSeq s
	val eq   = A.getIdIdent i
	val nb   = getNbOfId eq
	val sch1 =
	    case EN.getScheme eq (nb + 1) (getUnifEnv ()) of
		NONE => raise EH.Impossible ("Couldn't find " ^ eq ^ "'s environment")
	      | SOME scheme => scheme
	val t1   = (*E.mk_type_typ1*) E.mk_type_type D.dummy_label
	val typ  = foldr (fn (_, term) => E.mk_type_arrow (t1, term) D.dummy_label)
			t1
			(List.tabulate (ar, fn _ => ()))
	val sch2 = E.mk_new_scheme [] typ
	val _    = addOneId eq
	val context1 = add_param_context context eq sch1
	val context2 = add_param_context context1 tc sch2
	val context3 = add_new_id_context context2 eq
	val context4 = add_eqdec_context context3 tc eq
	val context5 = add_arity_context context4 eq eq []
    in ([], [], context5, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_TYCON),   label, value, regions, parents, children = [s, tc, ty]})      =
    let val id = A.getIdIdent tc
    in ([], [], rm_eqdec_context context id, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_EQTYCON), label, value, regions, parents, children = [s, tc, i, ty]})   =
    let val tc = A.getIdIdent tc
	val eq = A.getIdIdent i
	val _  = addOneId eq
	val context1 = add_new_id_context context eq
	val context2 = add_eqdec_context context1 tc eq
	val context3 = add_arity_context context2 eq eq []
    in ([], [], context3, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_TYFUN),   label, value, regions, parents, children = [seq, tc, ty]})    =
    ([], [], context, NONE)
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_MSGS),    label, value, regions, parents, children})                    =
    let val (decs, progs) = ListPair.unzip (map (toNuprlTAtom context) children)
    in (List.concat decs, List.concat progs, context, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_PARSE),   label, value, regions, parents, children = []})               =
    raise EH.Impossible "term is not parsable"
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_TYPEOF),  label, value, regions, parents, children = [id]})             =
    ([], [], context, NONE)
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_MAIN),    label, value, regions, parents, children = [exp]})            =
    let val (nt, pr) = toNuprlTerm context exp
	val name     = "main"
	val prod     = NT.mk_nuprl_prod_term NT.mk_nuprl_loc_term (NT.mk_nuprl_msg_term "i")
	val cls      = NT.mk_nuprl_class_term prod
	val (ab, th, rest, _) = genAbstractionAndWF 5 context value name nt exp (SOME (cls, [])) [] true false NONE
	val params   = getDecFP ()
	val _        = resetDecFP ()
    in (ab :: th :: rest, [], context, SOME (params, pr))
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_SPEC),    label, value, regions, parents, children = [id]})             =
    ([], [], context, NONE)
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_IMPORT),  label, value, regions, parents, children})                    =
    ([], [], context, NONE)
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_DOC),     label, value, regions, parents, children})                    =
    ([], [], context, NONE)
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_INV),     label, value, regions, parents, children = [id, cls, params, prop]}) =
    let val name   = A.getIdIdent id
	val ncls   = A.getIdIdent cls
	val prms   = A.getChildren params
	val nargs  = map (fn x => #1 (toNuprlTerm context x)) prms
	val nprop  = #1 (toNuprlTerm context prop)
	val (x,B)  = NT.dest_lambda nprop
	val nprop' = NT.mk_nuprl_lambda_term x (NT.mk_nuprl_assert_term B)
	val typ    =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE => raise EH.Impossible "toNuprlDec:no-type-for-invariant"
	      | SOME ity => toNuprlIty_em 16 true context (E.stripIty ity)
	val thm    = genInvariant context name ncls (nargs, prms) (nprop', prop) typ
	val _      = resetDecFP ()
    in ([thm], [], context, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_ORDER),   label, value, regions, parents, children = [id, cls, params, prop]}) =
    let val name   = A.getIdIdent id
	val ncls   = A.getIdIdent cls
	val prms   = A.getChildren params
	val nargs  = map (fn x => #1 (toNuprlTerm context x)) prms
	val nprop  = #1 (toNuprlTerm context prop)
	val (x,B') = NT.dest_lambda nprop
	val (y,B)  = NT.dest_lambda B'
	val assert = NT.mk_nuprl_assert_term B
	val lam    = NT.mk_nuprl_lambda_term y assert
	val nprop' = NT.mk_nuprl_lambda_term x lam
	val typ    =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE => raise EH.Impossible "toNuprlDec:no-type-for-invariant"
	      | SOME ity => toNuprlIty_em 16 true context (E.stripIty ity)
	val thm    = genInvariant context name ncls (nargs, prms) (nprop', prop) typ
	val _      = resetDecFP ()
    in ([thm], [], context, NONE)
    end
  | toNuprlDec context doc (term as A.N {kind = (A.DEC, A.DEC_DATA),    label, value, regions, parents, children = (id :: cons)}) =
    let val name  = A.getIdIdent id
	val data' =
	    foldr (fn (con, disju) =>
		      let val lab = A.getLabel con
			  val typ =
			      case EN.getItyAtLab (getUnifEnv ()) lab of
				  NONE     => raise EH.Impossible "We should have generated a type for this message"
				| SOME ity => toNuprlIty 18 false context (IDENTS.singleton name) (E.stripIty ity) BTV.empty E.emptyETVS
		      in case disju of
			     NONE => SOME typ
			   | SOME t => SOME (NT.mk_nuprl_union_term typ t)
		      end)
		  NONE
		  cons
	(* - the types of the get_ functions of the cons have to be of the form is_ => get_.
	 * - we need to fix the types of data's wf lemma. *)
	val data =
	    case data' of
		NONE => raise Fail "datatype"
	      | SOME t => NT.mk_nuprl_rec_term name t
	val (ab, th, rest, context') = genAbstractionAndWF 5 context "" name data dummy_dot_term (SOME (NT.mk_nuprl_type_term "i", [])) [] false true NONE
	val _ = resetDecFP ()
	val len = List.length cons
	val (decs, progs, _) =
	    foldl (fn (t, (decs, progs, n)) =>
		      let val (decs', progs') = toNuprlData context' name n len t
		      in (decs @ decs', progs @ progs', n + 1)
		      end)
		  ([], [], 0)
		  cons
    in (ab :: th :: rest @ decs, progs, context', NONE)
    end
  | toNuprlDec context doc term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun generateHeaders () =
    let val hdrs  = getHeaders ()
    in if not (null hdrs)
       then let val name_all      = "headers"
		val name_noinputs = "headers_no_inputs"
		val n_id_all      = mk_nuprl_user_term name_all      []
		val n_id_noinputs = mk_nuprl_user_term name_noinputs []
		val (n_lst_all, n_lst_noinputs) =
		    foldr (fn ((kind, hdr), (lst_all, lst_noinputs)) =>
			      let val lst_all' = NT.mk_nuprl_cons_term hdr lst_all
				  val lst_noinputs' =
				      if kind = "input"
				      then lst_noinputs
				      else NT.mk_nuprl_cons_term hdr lst_noinputs
			      in (lst_all', lst_noinputs')
			      end)
			  (NT.mk_nuprl_nil_term, NT.mk_nuprl_nil_term)
			  hdrs
		val n_typ = NT.mk_nuprl_list_term NT.mk_nuprl_name_term
		fun gen name id lst =
		    let val n_abs = NT.mk_nuprl_iabstraction_term id lst
			val n_mem = NT.mk_nuprl_member_term n_typ id
			val n_thm = NT.mk_nuprl_itheorem_term (get_user_name name ^ "_wf") n_mem
		    in [n_abs, n_thm]
		    end
		val objs1 = gen name_all      n_id_all      n_lst_all
		val objs2 = gen name_noinputs n_id_noinputs n_lst_noinputs
		val _     = resetHeaders ()
	    in SOME (n_id_noinputs, objs1 @ objs2)
	    end
       else NONE
    end

fun getHeadersObjs (SOME (_, objs)) = objs
  | getHeadersObjs NONE = []

fun getHeadersNoInputs (SOME (id, _)) = id
  | getHeadersNoInputs NONE = NT.mk_nuprl_nil_term

fun generateStdMa context id term =
    (case A.getKind term of
	 (A.DEC, A.DEC_MAIN) =>
	 (case A.getChildren term of
	      [_] =>
	      let val name_stdma      = "stdma"
		  val name_constraint = "message-constraint"
		  val name_delivered  = "messages-delivered"
		  val exp             = A.idToExp "main"
		  val (nt, pr)        = toNuprlTerm context exp
		  val set             = get_all_ids_list [nt]
		  val eo              = A.newIdIdSet set "eo"
		  val neo             = NT.mk_nuprl_variable_term eo
		  val teo             = NT.mk_nuprl_event_ordering_p_term
		  val propip          = NT.mk_nuprl_prop_term "i'"
		  val nity            = NT.mk_nuprl_fun_term teo propip
		  val lvl             = SOME "i"
		  val stdma           = NT.mk_nuprl_std_ma_term             neo nt id
		  val constraint      = NT.mk_nuprl_message_constraint_term neo nt id
		  val delivered       = NT.mk_nuprl_messages_delivered_term neo nt
		  val lstdma          = NT.mk_nuprl_lambda_term eo stdma
		  val lconstraint     = NT.mk_nuprl_lambda_term eo constraint
		  val ldelivered      = NT.mk_nuprl_lambda_term eo delivered
		  fun toAbs n name nterm =
		      let val (ab, th, rest, _) = genAbstractionAndWF n context "" name nterm exp (SOME (nity, [])) [] false false lvl
		      in ab :: th :: rest
		      end
		  val res =
		      toAbs 6 name_stdma      lstdma      @
		      toAbs 7 name_constraint lconstraint @
		      toAbs 8 name_delivered  ldelivered
		  val _ = resetDecFP ()
	      in res
	      end
	    | _ => [])
       | (A.DEC, _) => []
       | _ => List.concat (map (generateStdMa context id) (A.getChildren term)))

fun getSpecPrefix base addpref [] =
    if addpref
    then (base, [])
    else ("", [])
  | getSpecPrefix base addpref (lst as (spec :: terms)) =
    if A.getKind spec = (A.DEC, A.DEC_SPEC)
    then case A.getChildren spec of
	     [id] => (A.getIdIdent id, terms)
	   | _ => raise EH.Impossible A.wrongFormat
    else if addpref
    then (base, lst)
    else ("", lst)

fun mk_comment pref =
    let val name  = "------" ^ " " ^ pref ^ " - extra " ^ "------"
	val body  = NT.mk_nuprl_itext_nil_term
    in NT.mk_nuprl_icomment_term name body
    end

fun get_prog_decs (SOME prog) NONE = SOME prog
  | get_prog_decs _ (SOME prog) = SOME prog
  | get_prog_decs NONE NONE = NONE

fun toNuprlDecs context [] = ([], [], context, NONE)
  | toNuprlDecs context ((A.N {kind = (A.DEC, A.DEC_DOC), label, value, regions, parents, children}) :: dec :: decs) =
    let val doc = map A.getValue children
	val (terms1, progs1, context1, prog1) = toNuprlDec  context doc dec
	val (terms2, progs2, context2, prog2) = toNuprlDecs context1 decs
    in (terms1 @ terms2, progs1 @ progs2, context2, get_prog_decs prog1 prog2)
    end
  | toNuprlDecs context (dec :: decs) =
    let val (terms1, progs1, context1, prog1) = toNuprlDec  context [] dec
	val (terms2, progs2, context2, prog2) = toNuprlDecs context1 decs
    in (terms1 @ terms2, progs1 @ progs2, context2, get_prog_decs prog1 prog2)
    end

(* FILE *)
fun toNuprlFile fbase addpref context (term as A.N {kind = (A.FILE, A.FILE_F), label, value, regions, parents, children}) =
    let val (pref, children') = getSpecPrefix fbase addpref children
	val _     = setPrefix pref
	val (terms, progs, context', prog) = toNuprlDecs context children'
	val com   = mk_comment pref
	val hop   = generateHeaders ()
	val hdrs  = getHeadersObjs hop
	val stdma = generateStdMa context' (getHeadersNoInputs hop) term
	val _     = resetPrefix ()
	val extra = hdrs @ stdma
    in (terms @ (if List.null extra then [] else [com]) @ extra, progs, context', prog)
    end
  | toNuprlFile fbase addpref context term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun toNuprlFiles fbase addpref context [] = ([], [], context, NONE)
  | toNuprlFiles fbase addpref context (file :: files) =
    let val (terms1, progs1, context1, prog1) = toNuprlFile  fbase addpref context  file
	val (terms2, progs2, context2, prog2) = toNuprlFiles fbase addpref context1 files
    in (terms1 @ terms2, progs1 @ progs2, context2, get_prog_decs prog1 prog2)
    end

(* PROG *)
fun toNuprlProg fbase addpref context (term as A.N {kind = (A.PROG, A.PROG_P), label, value, regions, parents, children}) =
    toNuprlFiles fbase addpref context children
  | toNuprlProg fbase addpref context term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun generateProperties fname =
    let val st1   = NT.mk_nuprl_istring_term fname
	val prop1 = NT.mk_nuprl_iproperty_term fname st1
	val cons1 = NT.mk_nuprl_icons_cons_term prop1 NT.mk_nuprl_icons_nil_term
    in cons1
    end

fun loadNuprlPrelude () =
    let val lst = [("inl",  "inl",  [("x", [])]),
		   ("inr",  "inr",  [("x", [])]),
		   ("isl",  "isl",  [("x", [])]),
		   ("!",    "bnot", [("x", [])]),
		   ("fst",  "pi1",  [("x", [])]),
		   ("snd",  "pi2",  [("x", [])]),
		   ("eqof", "eqof", [("x", [])])]
    in foldl (fn ((id, name, ar), context) =>
		 add_arity_context context id name ar)
	     mk_empty_context
	     lst
    end

fun replace_prog (id, p) (NT.TERM (operator as ((opid, tag), params), bterms)) =
    if id = opid
    then p
    else let val bterms' =
		 map (fn (NT.B_TERM (vars, term)) =>
			 NT.B_TERM (vars, replace_prog (id, p) term))
		     bterms
	 in NT.TERM (operator, bterms')
	 end

fun form_prog' [] p = p
  | form_prog' (idp :: progs) p = replace_prog idp (form_prog' progs p)

fun form_prog context lst NONE = NONE
  | form_prog context lst (SOME (params, p)) =
    let val prog = form_prog' lst p
    (* It might be better to just get the parameters in the order in which
     * they occur in the spec. *)
	val lam = NT.mk_nuprl_lambdas_term params prog
    in SOME (params, lam)
    end

fun toStringProg NONE = "---"
  | toStringProg (SOME p) = NT.toStringTerm p

fun toNuprl mapop file obid term env sub addpref =
    let val _       = A.setIdsTerm term
	val _       = reset ()
	val _       = setObid obid
	val ctxt1   = loadNuprlPrelude ()
	val ctxt2   = add_alldefs_context ctxt1 mapop
	val solved  = EN.solver env sub
	val _       = setUnifEnv solved
	val fname   = #file (OS.Path.splitDirFile file)
	val fbase   = #base (OS.Path.splitBaseExt fname)
	val (terms, progs, context, prog) = toNuprlProg fbase addpref ctxt2 term
	val p       = form_prog context progs prog
	(*val _       = print (toStringProg p)*)
	val cons1   = generateProperties fname
	val n_terms = (*cons1 ::*) terms
	val _       = A.resetSimp ()
    in (n_terms, p)
    end


(* ------ EXPORT NUPRL TO E# ------ *)

(*fun prelude () =
    let val list_ind =
	    let val nilcase  = A.idToExp "nilcase"
		val patnil   = A.idToPat "nil"
		val branch1  = A.mk_new_term A.MRULE_M "" [] [patnil, nilcase]
		val conscase = A.idToExp "conscase"
		val x_id     = A.mk_new_term A.ID_VID "x"  [] []
		val xs_id    = A.mk_new_term A.ID_VID "xs" [] []
		val z_id     = A.mk_new_term A.ID_VID "z"  [] []
		val x_atexp  = A.mk_new_term A.ATEXP_ID "" [] [x_id]
		val xs_atexp = A.mk_new_term A.ATEXP_ID "" [] [xs_id]
		val z_atexp  = A.mk_new_term A.ATEXP_ID "" [] [z_id]
		val app1     = A.mk_new_term A.EXP_APP "" [] [conscase, x_atexp]
		val app2     = A.mk_new_term A.EXP_APP "" [] [app1, xs_atexp]
		val app3     = A.mk_new_term A.EXP_APP "" [] [app2, z_atexp]
		val lam      = A.mk_new_term A.EXP_LAMBDA "" [] [z_id, app3]
		val patcons  = A.mk_new_term A.PAT_CONS "." [] [A.idToPat "x", A.idToPat "xs"]
		val branch2  = A.mk_new_term A.MRULE_M "" [] [patcons, lam]
		val atwild   = A.mk_new_term A.ATPAT_WILD "" [] []
		val patwild  = A.mk_new_term A.PAT_ATPAT "" [] [atwild]
		(* NOTE: for the last branch, we don't case what's the expression,
		 * this branch is never gonna be reached. *)
		val branch3  = A.mk_new_term A.MRULE_M "" [] [patwild, nilcase]
		val match    = A.mk_new_term A.MATCH_M "" [] [branch1, branch2]
		val m_exp    = A.mk_new_term A.EXP_MATCH "" [] [A.idToExp "L", match]
		val L_id     = A.mk_new_term A.ID_VID "L" [] []
		val nil_id   = A.mk_new_term A.ID_VID "nilcase" [] []
		val cons_id  = A.mk_new_term A.ID_VID "conscase" [] []
		val L_arg    = A.mk_new_term A.ATPAT_ID "" [] [L_id]
		val nil_arg  = A.mk_new_term A.ATPAT_ID "" [] [nil_id]
		val cons_arg = A.mk_new_term A.ATPAT_ID "" [] [cons_id]
		val args     = A.mk_new_term A.PARAM_P "" [] [L_arg, nil_arg, cons_arg]
		val li_id    = A.mk_new_term A.ID_VID "list_ind" [] []
		val bind     = A.mk_new_term A.BIND_DEC "" [] [li_id, args, m_exp]
	    in A.mk_new_term A.DEC_LET "" [] [bind]
	    end
    in [list_ind]
    end*)

fun fromNuprlExp (NT.TERM ((("lambda", "OPID"), parameters), [NT.B_TERM ([var], term)])) =
    let val id  = A.mk_new_term A.ID_VID var [] []
	val exp = fromNuprlExp term
    in A.mk_new_term A.EXP_LAMBDA "" [] [id, exp]
    end
  | fromNuprlExp (NT.TERM ((("natural_number", "OPID"), [(n, "n")]), [])) =
    (case Int.fromString n of
	 NONE     => raise EH.Impossible "wrong format"
       | SOME nat =>
	 let val sc    = A.mk_new_term A.SCON_INT n [] []
	     val atexp = A.mk_new_term A.ATEXP_SCON "" [] [sc]
	 in A.mk_new_term A.EXP_ATEXP "" [] [atexp]
	 end)
  | fromNuprlExp (NT.TERM ((("variable", "OPID"), [(var, "v")]), [])) = A.idToExp var
  | fromNuprlExp (NT.TERM ((("apply", "OPID"), []), [NT.B_TERM ([], term1), NT.B_TERM ([], term2)])) =
    let val exp1  = fromNuprlExp term1
	val exp2  = fromNuprlExp term2
	val atexp = A.mk_new_term A.ATEXP_PAREN "" [] [exp2]
    in A.mk_new_term A.EXP_APP "" [] [exp1, atexp]
    end
  | fromNuprlExp (NT.TERM ((("list_ind", "OPID"), []), [NT.B_TERM ([], L), NT.B_TERM ([], nilcase), NT.B_TERM ([x, xs, r], conscase)])) =
    let val li_exp     = A.idToExp "list_ind"
	val L_exp      = fromNuprlExp L
	val L_atexp    = A.mk_new_term A.ATEXP_PAREN "" [] [L_exp]
	val nil_exp    = fromNuprlExp nilcase
	val nil_atexp  = A.mk_new_term A.ATEXP_PAREN "" [] [nil_exp]
	val x_id       = A.mk_new_term A.ID_VID x  [] []
	val xs_id      = A.mk_new_term A.ID_VID xs [] []
	val r_id       = A.mk_new_term A.ID_VID r  [] []
	val cons_exp   = fromNuprlExp conscase
	val lam1       = A.mk_new_term A.EXP_LAMBDA "" [] [r_id,  cons_exp]
	val lam2       = A.mk_new_term A.EXP_LAMBDA "" [] [xs_id, lam1]
	val lam3       = A.mk_new_term A.EXP_LAMBDA "" [] [x_id,  lam2]
	val lam_atexp  = A.mk_new_term A.ATEXP_PAREN "" [] [lam3]
	val app1       = A.mk_new_term A.EXP_APP "" [] [li_exp, L_atexp]
	val app2       = A.mk_new_term A.EXP_APP "" [] [app1, nil_atexp]
	val app3       = A.mk_new_term A.EXP_APP "" [] [app2, lam_atexp]
    in app3
    end
  (* TODO: Othersize we've got to look it up *)
  | fromNuprlExp _ = raise EH.Impossible "can't import the Nuprl term"

fun fromNuprlLhsRhs id lhs rhs =
    case lhs of
	NT.TERM (((opid, "OPID"), parameters), []) =>
	if opid = id
	then let val id   = A.mk_new_term A.ID_VID id [] []
		 val exp  = fromNuprlExp rhs
		 val list = map (fn (value, kind) =>
				    let val id    = A.mk_new_term A.ID_VID value [] []
				    in A.mk_new_term A.ATPAT_ID "" [] [id]
				    end)
				  parameters
		 val args = A.mk_new_term A.PARAM_P  "" [] list
		 val bind = A.mk_new_term A.BIND_DEC "" [] [id, args, exp]
		 val dec  = A.mk_new_term A.DEC_LET  "" [] [bind]
	     in dec
	     end
	else raise EH.Impossible "wrong term"
      | _ => raise EH.Impossible "wrong format"

end
